import logging

from django.db import migrations, transaction

logger = logging.getLogger(__name__)

COMPONENT_MAPPING = {
    "Internal Messaging": "Helpcenter after sales",
    "Crisis Communication Room": "Other",
    "data-visitors": "Data Ingestion",
    "data-buyers": "Data Ingestion",
    "Seller Order": "Seller Admin and Experience",
    "Legacy": "Other",
    "Authentification": "Customer login & signup",
    "data-operations": "Data Ingestion",
    "Test Incident": "Other",
    "Legal": "Other",
}

# Global variable to store incidents with the previous component
# This will be used to restore the original component in case of rollback
INCIDENTS_BACKUP = {}


def forwards_func(apps, _schema_editor):
    Incident = apps.get_model("incidents", "Incident")
    Component = apps.get_model("incidents", "Component")
    with transaction.atomic():
        for old_component_name, new_component_name in COMPONENT_MAPPING.items():
            old_component = Component.objects.filter(name=old_component_name).first()
            if not old_component:
                logger.error(f"Failed to find component {old_component_name}.")
                continue
            incidents = Incident.objects.filter(component=old_component)
            ids = list(incidents.values_list("id", flat=True))
            if ids:
                INCIDENTS_BACKUP[old_component_name] = ids
                new_component = Component.objects.filter(name=new_component_name).first()
                if not new_component:
                    logger.error(f"Failed to find component {new_component_name}.")
                    continue
                incidents.update(component=new_component)


def backwards_func(apps, _schema_editor):
    Incident = apps.get_model("incidents", "Incident")
    Component = apps.get_model("incidents", "Component")
    with transaction.atomic():
        for old_component_name, ids in INCIDENTS_BACKUP.items():
            old_component = Component.objects.filter(name=old_component_name).first()
            if not old_component:
                logger.error(f"Failed to find component {old_component_name}.")
                continue
            Incident.objects.filter(id__in=ids).update(component=old_component)


def delete_old_components(apps, _schema_editor):
    Component = apps.get_model("incidents", "Component")
    try:
        for old_name in COMPONENT_MAPPING:
            Component.objects.filter(name=old_name).delete()
    except Exception:
        logger.exception(f"Failed to delete old components {old_name}.")


def recreate_old_components(apps, _schema_editor):
    Component = apps.get_model("incidents", "Component")
    try:
        for old_name in COMPONENT_MAPPING:
            Component.objects.get_or_create(name=old_name)
    except Exception:
        logger.exception(f"Failed to recreate old components {old_name}.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0010_update_components"),
    ]

    operations = [
        migrations.RunPython(forwards_func, backwards_func),
        migrations.RunPython(
            delete_old_components,
            recreate_old_components,
        ),
    ]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from firefighter.slack.slack_app import SlackApp
from firefighter.slack.views.modals.base_modal.base import SlackModal
from firefighter.slack.views.modals.base_modal.base_mixins import (
    IncidentSelectableModalMixinBase,
)
from firefighter.slack.views.modals.select import modal_select

if TYPE_CHECKING:
    from re import Pattern

    from slack_sdk.models.views import View

app = SlackApp()

logger = logging.getLogger(__name__)


class IncidentSelectableModalMixin(IncidentSelectableModalMixinBase):
    callback_id: str | Pattern[str]
    select_modal_title: str = "Select an incident"
    select_incident_title: str = "Select a critical incident"

    def build_modal_with_context(self, body: dict[str, Any], **kwargs: Any) -> View:
        # XXX Should use a Protocol for typings
        if not isinstance(self, SlackModal):
            err_msg = (
                f"{self.__class__.__name__} must be a subclass of {SlackModal.__name__}"
            )
            raise TypeError(err_msg)
        self._get_kwargs_builder(body, kwargs, self.builder_fn_args)
        if "body" in self.builder_fn_args:
            kwargs["body"] = body
        elif "body" in kwargs:
            del kwargs["body"]

        if "incident" in self.builder_fn_args and kwargs.get("incident") is None:
            if kwargs.get("callback_id") is None:
                kwargs["callback_id"] = self.callback_id
            kwargs.pop("body", None)
            return modal_select.build_modal_fn(**kwargs, select_class=self)

        return self.build_modal_fn(**kwargs)  # type: ignore

    def get_callback_id(self) -> str | Pattern[str]:
        return self.callback_id

    def get_select_modal_title(self) -> str:
        return self.select_modal_title

    def get_select_title(self) -> str:
        return self.select_incident_title



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.blocks import ContextBlock

from firefighter.raid.forms import (
    CreateNormalCustomerIncidentForm,
    CreateRaidDocumentationRequestIncidentForm,
    CreateRaidFeatureRequestIncidentForm,
    CreateRaidInternalIncidentForm,
    RaidCreateIncidentSellerForm,
)
from firefighter.slack.views.modals.opening.set_details import SetIncidentDetails

if TYPE_CHECKING:
    from firefighter.slack.views.modals.base_modal.form_utils import (
        SlackFormAttributesDict,
    )

logger = logging.getLogger(__name__)

slack_fields: SlackFormAttributesDict = {
    "title": {
        "input": {
            "multiline": False,
            "placeholder": "Summary of the issue.",
        },
        "block": {"hint": None},
    },
    "description": {
        "input": {
            "multiline": True,
            "placeholder": "Explain your issue in English giving as much details as possible. It helps people handling the issue. \nThis description can be edited later.",
        },
        "block": {"hint": None},
    },
    "suggested_team_routing": {
        "widget": {
            "post_block": ContextBlock(
                elements=[
                    MarkdownTextObject(
                        text="Feature Team or Train that should own the issue. If you don't know access <https://manomano.atlassian.net/wiki/spaces/QRAFT/pages/3970335291/Teams+and+owners|here> for guidance."
                    ),
                ]
            )
        },
    },
    "area": {
        "input": {
            "placeholder": "Select affected area",
        },
    },
}


class CreateRaidCustomerIncidentFormSlack(CreateNormalCustomerIncidentForm):
    slack_fields: SlackFormAttributesDict = slack_fields


class OpeningRaidCustomerModal(SetIncidentDetails[CreateRaidCustomerIncidentFormSlack]):
    open_action: str = "open_incident_raid_customer_request"
    push_action: str = "push_raid_customer_request"
    callback_id: str = "open_incident_raid_customer_request"
    id = "raid_customer_request"
    title = "New Customer Incident"

    form_class = CreateRaidCustomerIncidentFormSlack


class CreateRaidDocumentationRequestIncidentFormSlack(
    CreateRaidDocumentationRequestIncidentForm
):
    slack_fields: SlackFormAttributesDict = slack_fields


class CreateRaidFeatureRequestIncidentFormSlack(CreateRaidFeatureRequestIncidentForm):
    slack_fields: SlackFormAttributesDict = slack_fields


class OpeningRaidFeatureRequestModal(
    SetIncidentDetails[CreateRaidFeatureRequestIncidentFormSlack]
):
    open_action: str = "open_incident_raid_feature_request"
    push_action: str = "push_raid_feature_request"
    callback_id: str = "open_incident_raid_feature_request"

    title = "New Feature Request"
    form_class = CreateRaidFeatureRequestIncidentFormSlack


class OpeningRaidDocumentationRequestModal(
    SetIncidentDetails[CreateRaidDocumentationRequestIncidentFormSlack]
):
    open_action: str = "open_incident_raid_documentation_request"
    push_action: str = "push_raid_documentation_request"
    callback_id: str = "open_incident_raid_documentation_request"

    title = "New Documentation Request"
    form_class = CreateRaidDocumentationRequestIncidentFormSlack


class CreateRaidInternalIncidentFormSlack(CreateRaidInternalIncidentForm):
    slack_fields: SlackFormAttributesDict = slack_fields


class OpeningRaidInternalModal(SetIncidentDetails[CreateRaidInternalIncidentFormSlack]):
    open_action: str = "open_incident_raid_internal_request"
    push_action: str = "push_raid_internal_request"
    callback_id: str = "open_incident_raid_internal_request"

    title = "New Internal Incident"

    form_class = CreateRaidInternalIncidentFormSlack


class CreateRaidSellerIncidentFormSlack(RaidCreateIncidentSellerForm):
    slack_fields: SlackFormAttributesDict = slack_fields


class OpeningRaidSellerModal(SetIncidentDetails[CreateRaidSellerIncidentFormSlack]):
    open_action: str = "open_incident_raid_seller_request"
    push_action: str = "push_raid_seller_request"
    callback_id: str = "open_incident_raid_seller_request"

    title = "New Seller Incident"
    form_class = CreateRaidSellerIncidentFormSlack


# Instantiate all the modals to register actions
_modals = [
    raid_seller := OpeningRaidSellerModal(),
    raid_internal := OpeningRaidInternalModal(),
    raid_customer := OpeningRaidCustomerModal(),
    raid_feature := OpeningRaidFeatureRequestModal(),
    raid_documentation := OpeningRaidDocumentationRequestModal(),
]



# Generated by ChatGPT on 2024-06-24
from django.db import migrations


def update_impactlevel_names(apps, schema_editor):
    ImpactType = apps.get_model("incidents", "ImpactType")
    ImpactLevel = apps.get_model("incidents", "ImpactLevel")

    # Map: (impact_type.value, impact_level.value) -> new name
    updates = {
        ("business_impact", "HT"): "Critical Impact (>5% BV loss)",
        ("business_impact", "HI"): "Major Impact (<5% BV loss)",
        ("business_impact", "MD"): "Uncertain Business Impact",
        ("business_impact", "LO"): "Low Impact not measurable",
        ("business_impact", "LT"): "Very Low Impact not measurable",
        ("customers_impact", "HT"): "Critical issue for many customers",
        ("customers_impact", "HI"): "Major issue for many customers",
        ("customers_impact", "MD"): "Major issue for few customers",
        ("customers_impact", "LO"): "Minor issue for customers",
        ("customers_impact", "LT"): "Cosmetic minor issue for customers",
        ("sellers_impact", "HT"): "Critical issue for many sellers",
        ("sellers_impact", "HI"): "Major issue for many sellers",
        ("sellers_impact", "MD"): "Major issue for few sellers",
        ("sellers_impact", "LO"): "Minor issue for sellers",
        ("sellers_impact", "LT"): "Cosmetic minor issue for sellers",
        ("employees_impact", "MD"): "Departments fully blocked",
        ("employees_impact", "LO"): "Some employees affected",
        ("employees_impact", "LT"): "Degraded service for employees",
        ("employees_impact", "NO"): "No impact for employees",
    }

    for (impact_type_value, level_value), new_name in updates.items():
        impact_type = ImpactType.objects.filter(value=impact_type_value).first()
        if not impact_type:
            continue
        impact_level = ImpactLevel.objects.filter(impact_type=impact_type, value=level_value).first()
        if impact_level:
            impact_level.name = new_name
            impact_level.save(update_fields=["name"])


class Migration(migrations.Migration):
    dependencies = [
        ("incidents", "0017_reorder_impact_types"),
    ]
    operations = [
        migrations.RunPython(update_impactlevel_names),
    ]



from __future__ import annotations

from pathlib import Path

from decouple import AutoConfig

BASE_DIR = Path(__file__).parent.parent.parent

# Loading `.env` files with decouple
config = AutoConfig(search_path=BASE_DIR)

env: str = config("ENV", default="dev", cast=str)
env = env.strip()
_ENV = SETTINGS_ENV = env

if _ENV.lower() in {"prod", "prd", "support", "int"}:
    SETTINGS_ENV = "prod"
ENV = SETTINGS_ENV

# Commonly reused params
FF_VERSION: str | int = config("VERSION", default="dev")
APP_DISPLAY_NAME: str = config("APP_DISPLAY_NAME", default="FireFighter")
"""The name of the app. Used in the title of the app, and in the navigation bar."""
BASE_URL: str = config("BASE_URL")
"""The base URL of the app. Used for links in externals surfaces, like Slack or documents."""



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings

FF_USER_ID_HEADER: str = settings.FF_USER_ID_HEADER

if TYPE_CHECKING:
    from collections.abc import Callable

    from django.http import HttpRequest, HttpResponse


class HeaderUser:
    """Adds the user ID to the response headers configured with "FF-User-Id", to log in access logs."""

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request: HttpRequest) -> HttpResponse:
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        request_user_id = request.user.id if request.user.is_authenticated else None

        response: HttpResponse = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        if request_user_id:
            response.headers[FF_USER_ID_HEADER] = str(request_user_id)

        return response



from __future__ import annotations

from django.contrib import admin

from firefighter.jira_app.models import JiraIssue, JiraUser


@admin.register(JiraUser)
class JiraUserAdmin(admin.ModelAdmin[JiraUser]):
    model = JiraUser
    list_display = ["id", "user"]
    list_display_links = ["id", "user"]
    search_fields = ["id", "user__username"]
    list_select_related = ["user"]


@admin.register(JiraIssue)
class JiraIssueAdmin(admin.ModelAdmin[JiraIssue]):
    model = JiraIssue
    list_display = ["id", "key", "summary"]
    list_display_links = ["id", "key", "summary"]
    search_fields = ["id", "key", "summary", "description"]

    autocomplete_fields = ["watchers", "assignee", "reporter"]



from __future__ import annotations

import contextlib
import enum
from typing import TYPE_CHECKING
from urllib.parse import urlencode

from django.apps import apps
from django.conf import settings
from django.urls import NoReverseMatch, reverse
from simple_menu import Menu, MenuItem

if TYPE_CHECKING:
    from django.http import HttpRequest

    from firefighter.firefighter.utils import AuthenticatedHttpRequest

APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME


class Menus(enum.StrEnum):
    main = "main"
    footer = "footer"


def user_details_url(request: AuthenticatedHttpRequest) -> str | None:
    """Return a personalized title for our profile menu item."""
    if request.user.is_authenticated:
        return request.user.get_absolute_url()
    return None  # type: ignore[unreachable]


def log_out_url(_request: HttpRequest) -> str:
    url = reverse("oidc_logout")
    return "?".join([
        url,
        urlencode({"next": "/admin/login/", "fail": "/admin/login/"}),
    ])


def setup_navbar_menu() -> None:
    submenu_items = [
        MenuItem(
            "Declare a critical incident",
            reverse("incidents:incident-create"),
            boost=False,
        ),
        MenuItem(
            "Statistics",
            reverse("incidents:incident-statistics"),
        ),
        MenuItem(
            "Issue categories",
            reverse("incidents:component-list"),
        ),
    ]
    Menu.add_item(
        Menus.main,
        MenuItem(
            "Incidents",
            reverse(
                "incidents:incident-list",
            ),
            icon="menu-app",
            children=submenu_items,
        ),
    )

    if apps.is_installed("firefighter.pagerduty"):
        Menu.add_item(
            Menus.main,
            MenuItem(
                "On-calls",
                reverse("pagerduty:oncall-list"),
            ),
        )

    if apps.is_installed("firefighter.confluence"):
        knowledge_base_children = [
            MenuItem(
                "Runbooks",
                reverse("confluence:runbook_list"),
            ),
        ]
    else:
        knowledge_base_children = []
    Menu.add_item(
        Menus.main,
        MenuItem(
            "Knowledge Base",
            None,
            children=[
                *knowledge_base_children,
                MenuItem("Metrics Explanation", reverse("incidents:docs-metrics")),
                MenuItem("Role Types", reverse("incidents:docs-role-type-list")),
            ],
        ),
    )

    Menu.add_item(
        "user",
        MenuItem(
            "Profile",
            user_details_url,
            badge="Beta",
        ),
    )
    Menu.add_item(
        "user",
        MenuItem(
            "Back-office",
            reverse("admin:index"),
            check=lambda r: r.user.is_staff,
        ),
    )

    Menu.add_item(
        "user",
        MenuItem(
            "Log out",
            url=log_out_url,
            check=lambda r: r.user.is_authenticated,
        ),
    )


def setup_footer_menu() -> None:
    # XXX Link to FireFighter Github and documentation in default footer
    Menu.add_item(
        Menus.footer,
        MenuItem(
            APP_DISPLAY_NAME,
            None,
            children=[
                MenuItem("Github", "https://github.com/ManoManoTech/firefighter"),
            ],
        ),
    )

    with contextlib.suppress(NoReverseMatch):
        Menu.add_item(
            Menus.footer,
            MenuItem(
                f"{APP_DISPLAY_NAME} API",
                None,
                children=[
                    MenuItem("OpenAPI specs", reverse("api:schema")),
                    MenuItem("OpenAPI Swagger-UI", reverse("api:swagger-ui")),
                ],
            ),
        )


if not settings.FF_OVERRIDE_MENUS_CREATION:
    setup_navbar_menu()
    setup_footer_menu()
elif callable(settings.FF_OVERRIDE_MENUS_CREATION):
    settings.FF_OVERRIDE_MENUS_CREATION()
else:
    raise TypeError(
        "FF_OVERRIDE_MENUS_CREATION is not callable. "
        "Please provide a callable to override the menus creation."
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django import forms
from django.conf import settings

from firefighter.pagerduty.models import PagerDutyService

if TYPE_CHECKING:
    from django.db.models import QuerySet


class CreatePagerDutyIncidentFreeForm(forms.Form):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_service_queryset() -> QuerySet[PagerDutyService]:
        if settings.DEBUG:
            return PagerDutyService.objects.all().order_by("summary")
        return PagerDutyService.objects.filter(ignore=False).order_by("summary")

    title = forms.CharField(
        label="What's going on?",
        max_length=128,
        min_length=10,
        widget=forms.TextInput,
    )
    service = forms.ModelChoiceField(
        queryset=get_service_queryset(),
    )
    #  Assign to users / escalation policies

    urgency = forms.ChoiceField(choices=[("high", "High"), ("low", "Low")])  # TODO Enum
    details = forms.CharField(
        label="More details",
        widget=forms.Textarea(
            attrs={"rows": 5},
        ),
        min_length=10,
        max_length=1200,
    )



from __future__ import annotations

from firefighter.api.serializers import GroupSerializer
from firefighter.api.views._base import ReadOnlyModelViewSet
from firefighter.incidents.models.group import Group


class GroupViewSet(ReadOnlyModelViewSet[Group]):
    queryset = Group.objects.all()
    serializer_class = GroupSerializer



from __future__ import annotations

from firefighter.api.serializers import EnvironmentSerializer
from firefighter.api.views._base import ReadOnlyModelViewSet
from firefighter.incidents.models.environment import Environment


class EnvironmentViewSet(ReadOnlyModelViewSet[Environment]):
    queryset = Environment.objects.all()
    serializer_class = EnvironmentSerializer



# Generated by Django 4.2.21 on 2025-06-02 16:29
import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0011_update_incidents"),
    ]

    operations = [
        migrations.AlterField(
            model_name="impactlevel",
            name="id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="impactlevel",
            name="value",
            field=models.CharField(
                choices=[
                    ("HT", "Highest"),
                    ("HI", "High"),
                    ("MD", "Medium"),
                    ("LO", "Low"),
                    ("LT", "Lowest"),
                    ("NO", "N/A"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
        migrations.AlterField(
            model_name="impacttype",
            name="emoji",
            field=models.CharField(default="", max_length=5),
        ),
    ]



from __future__ import annotations

from django.db import models
from django_stubs_ext.db.models import TypedModelMeta


class IncidentCostType(models.Model):
    """Incident Cost Type is inspired from dispatch."""

    id = models.BigAutoField(primary_key=True)

    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(null=True, blank=True)
    category = models.CharField(max_length=255, null=True, blank=True)

    default = models.BooleanField(default=False)
    editable = models.BooleanField(default=True)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta(TypedModelMeta):
        verbose_name = "Incident cost type"
        verbose_name_plural = "Incident cost types"

    def __str__(self) -> str:
        return self.name



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django import forms
from django.forms.models import ModelChoiceIterator
from slack_sdk.errors import SlackApiError
from slack_sdk.models.blocks.blocks import Block, SectionBlock
from slack_sdk.models.views import View

from firefighter.slack.forms.sos_form import SosForm
from firefighter.slack.slack_templating import slack_block_footer, slack_block_separator
from firefighter.slack.views.modals.base_modal.base import ModalForm
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.user import User
    from firefighter.slack.models.sos import Sos
    from firefighter.slack.views.modals.base_modal.form_utils import (
        SlackFormAttributesDict,
    )

logger = logging.getLogger(__name__)


def sos_label(obj: Sos) -> str:
    return f"{obj.name}  ({obj.usergroup_slack_fmt} in #{obj.conversation.name})"


class SendSosFormSlack(SosForm):
    slack_fields: SlackFormAttributesDict = {
        "sos": {
            "input": {
                "placeholder": "Select your SOS",
            },
            "widget": {
                "label_from_instance": sos_label,
            },
        },
    }


class SendSosModal(IncidentSelectableModalMixin, ModalForm[SendSosFormSlack]):
    open_action: str = "open_modal_incident_send_sos"

    update_action: str = "update_modal_incident_send_sos"
    push_action: str = "push_modal_incident_send_sos"
    callback_id: str = "incident_send_sos"

    form_class = SendSosFormSlack

    def build_modal_fn(self, incident: Incident, **kwargs: Any) -> View:
        blocks: list[Block] = [
            SectionBlock(
                text=":bulb: You are about to request help from SREs, in their dedicated Slack channel.\nChoose the right train to alert the right people."
            )
        ]
        sos_form = self.get_form_class()()
        sos_field = sos_form.form.fields["sos"]
        if not isinstance(sos_field, forms.ModelChoiceField):
            raise TypeError("sos field must be a ModelChoiceField")
        sos_field_choices = sos_field.choices
        if not isinstance(sos_field_choices, list | ModelChoiceIterator):
            raise TypeError("sos field choices must be a list or a ModelChoiceIterator")
        has_sos_options = len(sos_field_choices) > 0
        if has_sos_options:
            blocks.extend(sos_form.slack_blocks())
        else:
            blocks.append(
                SectionBlock(
                    text=":warning: No SOS configured in the database! :warning:\nAdministrator action is needed. Please raise the issue."
                )
            )
            logger.error(
                "No SOS configured in the database! Please SOSes in the back-office."
            )
        blocks.extend((slack_block_separator(), slack_block_footer()))
        return View(
            type="modal",
            title=f"Get help for #{incident.id}"[:24],
            submit="Send SOS"[:24] if has_sos_options else None,
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    def handle_modal_fn(  # type: ignore
        self, ack: Ack, body: dict[str, Any], incident: Incident, user: User
    ):
        slack_form = self.handle_form_errors(
            ack, body, forms_kwargs={}, ack_on_success=False
        )
        if slack_form is None:
            return
        form: SendSosFormSlack = slack_form.form
        if len(form.cleaned_data) == 0:
            # XXX We should have a prompt for empty forms
            return
        try:
            self._send_incident_sos(incident, form.cleaned_data["sos"], user)
            ack()
        except SlackApiError:
            logger.exception(f"Error sending SOS for #{incident.id}")
            # XXX Better strategy for errors?
            ack()

    def get_select_modal_title(self) -> str:
        return "Send SOS"

    def get_select_title(self) -> str:
        return "Select a critical incident for which you need help"

    @staticmethod
    def _send_incident_sos(incident: Incident, sos: Sos, user: User) -> None:
        from firefighter.slack.messages.slack_messages import (  # noqa: PLC0415
            SlackMessagesSOS,
        )

        sos.conversation.send_message_and_save(
            SlackMessagesSOS(incident=incident, usergroup=sos.usergroup_slack_fmt)
        )
        incident.create_incident_update(
            message=f"Asked for SRE help in #{sos.conversation.name}",
            event_type="send_sos",
            created_by=user,
        )


modal_send_sos = SendSosModal()



from __future__ import annotations

import logging

import pytest
from pytest_mock import MockerFixture

from firefighter.incidents.factories import IncidentFactory
from firefighter.incidents.models import Incident, IncidentUpdate
from firefighter.slack.views.modals.status import StatusModal

logger = logging.getLogger(__name__)


@pytest.mark.django_db
class TestStatusModal:
    @staticmethod
    @pytest.fixture
    def incident() -> Incident:
        """Returns a valid incident."""
        return IncidentFactory.build()

    @staticmethod
    def test_status_modal_build(incident: Incident, mocker: MockerFixture) -> None:
        mocker.patch.object(
            StatusModal,
            "get_latest_update",
            return_value=IncidentUpdate(incident=incident, message="test"),
        )
        modal = StatusModal()
        res = modal.build_modal_fn(incident=incident)

        # To dict will validate the JSON as well (e.g. check the view title length)
        # Any validation error will raise an exception
        assert res.to_dict()

        values = res.to_dict()
        assert "blocks" in values
        assert len(values["blocks"]) == 7

    @staticmethod
    def test_status_modal_build_no_message(
        incident: Incident, mocker: MockerFixture
    ) -> None:
        mocker.patch.object(
            StatusModal,
            "get_latest_update",
            return_value=IncidentUpdate(incident=incident),
        )
        modal = StatusModal()
        res = modal.build_modal_fn(incident=incident)

        # To dict will validate the JSON as well (e.g. check the view title length)
        # Any validation error will raise an exception
        assert res.to_dict()

        values = res.to_dict()
        assert "blocks" in values
        assert len(values["blocks"]) == 4

    @staticmethod
    def test_status_modal_build_no_update(
        incident: Incident, mocker: MockerFixture
    ) -> None:
        mocker.patch.object(
            StatusModal,
            "get_latest_update",
            return_value=None,
        )
        modal = StatusModal()
        with pytest.raises(ValueError, match="No update found for incident"):
            modal.build_modal_fn(incident=incident)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.apps import apps
from slack_sdk.models.blocks.blocks import Block, SectionBlock
from slack_sdk.models.views import View

from firefighter.slack.views.modals.base_modal.base import SlackModal
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)

if apps.is_installed("firefighter.confluence"):
    from firefighter.confluence.models import PostMortem

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident


logger = logging.getLogger(__name__)


class PostMortemModal(
    IncidentSelectableModalMixin,
    SlackModal,
):
    open_action: str = "open_modal_create_postmortem"
    push_action: str = "push_modal_create_postmortem"
    callback_id: str = "incident_create_postmortem"

    def build_modal_fn(self, incident: Incident, **kwargs: Any) -> View:
        blocks: list[Block] = []
        if hasattr(incident, "postmortem_for"):
            blocks.extend([
                SectionBlock(
                    text=f"Postmortem for incident #{incident.id} has already been created."
                ),
                SectionBlock(
                    text=f"See the postmortem page <{incident.postmortem_for.page_url}|on Confluence>."
                ),
            ])
            submit = None
        else:
            blocks.extend([
                SectionBlock(
                    text=f"Postmortem does not yet exist for incident #{incident.id}."
                ),
                SectionBlock(
                    text="Click on the button to create the postmortem on Confluence."
                ),
            ])
            submit = "Create postmortem"[:24]

        return View(
            type="modal",
            title="Postmortem"[:24],
            submit=submit,
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    @staticmethod
    def handle_modal_fn(ack: Ack, incident: Incident) -> None:  # type: ignore[override]
        if not apps.is_installed("firefighter.confluence"):
            ack(text="Confluence is not enabled!")
            return
        if hasattr(incident, "postmortem_for"):
            ack(text="Post-mortem has already been created.")
            return
        ack()
        PostMortem.objects.create_postmortem_for_incident(incident)


modal_postmortem = PostMortemModal()



from __future__ import annotations

import pytest


def test_json_logging(caplog: pytest.LogCaptureFixture, capsys, settings) -> None:
    """This test ensures that JSON logging is working."""
    # ruff: noqa: PLC0415
    import logging

    from firefighter.firefighter.settings.components.logging import (
        get_json_formatter,
    )

    caplog.set_level(logging.INFO)
    # Override the current logging settings
    settings.LOGGING["formatters"]["dynamicfmt"] = get_json_formatter()

    from logging import config

    config.dictConfig(settings.LOGGING)
    logger = logging.getLogger(__name__)

    logger.info("Testing now.")

    captured = capsys.readouterr()
    assert '{"message": "Testing now.",' in captured.err
    # XXX Make more testing assertion about our expected JSON log output and its format.
    # This is a bit tricky as it is in Gunicorn logs.



from __future__ import annotations


class IncidentSelectableModalMixinBase:
    def get_select_title(self) -> str:
        raise NotImplementedError

    def get_select_modal_title(self) -> str:
        raise NotImplementedError



# Generated by Django 4.2.7 on 2023-12-04 11:09

import datetime
import uuid

import django.contrib.auth.validators
import django.core.validators
import django.db.models.deletion
import django.utils.timezone
import taggit.managers
from django.conf import settings
from django.db import migrations, models

import firefighter.incidents.models.priority
import firefighter.incidents.models.severity
import firefighter.incidents.models.user


class Migration(migrations.Migration):
    replaces = [
        ("incidents", "0001_initial"),
        ("incidents", "0002_incidentupdate_component"),
        ("incidents", "0003_alter_incidentupdate_severity"),
        ("incidents", "0004_auto_20210607_1053"),
        ("incidents", "0005_alter_incidentupdate_communication_lead"),
        ("incidents", "0006_alter_user_slack_id"),
        ("incidents", "0007_alter_incident_id"),
        ("incidents", "0008_auto_20210624_1135"),
        ("incidents", "0009_auto_20210625_1721_squashed_0014_alter_severity_value"),
        ("incidents", "0010_auto_20210627_2155"),
        ("incidents", "0011_auto_20210628_1427"),
        ("incidents", "0012_auto_20210702_1634"),
        (
            "incidents",
            "0013_environment_value_squashed_0017_rename_slack_groups_component_slack_conversations",
        ),
        ("incidents", "0014_auto_20210809_1052"),
        ("incidents", "0015_auto_20210820_1534"),
        ("incidents", "0016_auto_20210825_1039"),
        ("incidents", "0017_auto_20210829_1608"),
        ("incidents", "0018_alter_incidentupdate__status"),
        ("incidents", "0019_auto_20210903_1410"),
        ("incidents", "0020_auto_20211104_0951"),
        ("incidents", "0021_auto_20211112_1038"),
        ("incidents", "0022_alter_user_name"),
        ("incidents", "0023_auto_20220622_1559"),
        ("incidents", "0024_alter_incidentmembership_inviter"),
        ("incidents", "0025_alter_incident_commander_and_more"),
        ("incidents", "0026_alter_user_options"),
        ("incidents", "0027_component_private"),
        ("incidents", "0028_incident_private_alter_component_order_and_more"),
        ("incidents", "0027_remove_component_slack_conversations"),
        ("incidents", "0029_merge_20221031_1156"),
        ("incidents", "0030_status_enum_constraints"),
        ("incidents", "0031_incidentcosttype_incidentcost"),
        ("incidents", "0032_alter_incidentcost_options_and_more"),
        ("incidents", "0033_user_bot"),
        ("incidents", "0034_component_deploy_warning"),
        ("incidents", "0035_user_avatar"),
        ("incidents", "0036_severity_enabled_create_severity_enabled_update_and_more"),
        ("incidents", "0037_priority_migration"),
        ("incidents", "0038_alter_incident_severity"),
        (
            "incidents",
            "0039_impact_impacttype_incidentimpact_impact_impact_type_and_more",
        ),
        ("incidents", "0040_priority_recommended_response_type"),
        ("incidents", "0041_impactlevel_and_more"),
        ("incidents", "0042_alter_impactlevel_id"),
        ("incidents", "0043_impactlevel_emoji_impacttype_emoji"),
        ("incidents", "0044_dynamic_metrics"),
        ("incidents", "0045_dynamic_incident_roles"),
        ("incidents", "0046_dynamic_incident_roles_bis"),
        ("incidents", "0047_incidentroletype_incidents_incidentroletype_slug_check"),
        ("incidents", "0048_remove_incidentroletype_best_practices_and_more"),
        ("incidents", "0049_incidentrole_created_at_incidentrole_updated_at"),
        ("incidents", "0050_remove_incidentrole_unique_incident_role_and_more"),
    ]

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("taggit", "0003_taggeditem_add_unique_index"),
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="Component",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("slack_groups", models.CharField(default="", max_length=128)),
                ("cachet_id", models.IntegerField(null=True)),
                ("on_call", models.SmallIntegerField(default=0)),
                ("description", models.TextField()),
                ("order", models.IntegerField(default=0)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "ordering": ["order"],
            },
        ),
        migrations.CreateModel(
            name="Environment",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                ("description", models.CharField(max_length=128, unique=True)),
                ("order", models.IntegerField(default=0)),
                ("default", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("value", models.CharField(max_length=16, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                ("description", models.TextField(blank=True)),
                ("order", models.IntegerField(default=0, unique=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Severity",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(default="SEVXX", max_length=128, unique=True),
                ),
                ("emoji", models.CharField(default="🔥", max_length=5)),
                ("description", models.CharField(max_length=128)),
                ("order", models.IntegerField(default=0)),
                ("default", models.BooleanField(default=False)),
                ("needs_postmortem", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("value", models.IntegerField(unique=True)),
                (
                    "reminder_time",
                    models.DurationField(
                        default=datetime.timedelta(seconds=3600),
                        help_text="Time before an incident without any incident update will receive a reminder.",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "severities",
                "ordering": ["order"],
            },
        ),
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="User fullname. Prefer to set `first_name` and `last_name`, and get `full_name`. ",
                        max_length=128,
                    ),
                ),
                ("email", models.EmailField(max_length=128, unique=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="first name"
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, max_length=150, verbose_name="last name"
                    ),
                ),
                (
                    "password",
                    models.CharField(
                        default="!0PRxvh7bM2Twf93xctRExwjMqbaOnK2VaVkKiUzu",
                        max_length=128,
                        verbose_name="password",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={
                            "unique": "A user with that username already exists."
                        },
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[
                            django.contrib.auth.validators.UnicodeUsernameValidator()
                        ],
                        verbose_name="username",
                    ),
                ),
            ],
            options={
                "ordering": ["username"],
            },
            managers=[
                ("objects", firefighter.incidents.models.user.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name="Incident",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("title", models.CharField(max_length=128)),
                ("description", models.TextField()),
                (
                    "_status",
                    models.IntegerField(
                        choices=[
                            (10, "Open"),
                            (20, "Investigating"),
                            (30, "Mitigating"),
                            (40, "Mitigated"),
                            (50, "Post-mortem"),
                            (60, "Closed"),
                        ],
                        db_column="status",
                        default=10,
                        verbose_name="Status",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("closed_at", models.DateTimeField(blank=True, null=True)),
                (
                    "commander",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="incidents_commander",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "communication_lead",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="incidents_communication_lead",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "component",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="incidents.component",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "environment",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="incidents.environment",
                    ),
                ),
                (
                    "severity",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="incidents.severity",
                    ),
                ),
                (
                    "time_to_detect",
                    models.DurationField(
                        blank=True, db_column="time_to_detect", null=True
                    ),
                ),
                (
                    "time_to_internal_response",
                    models.DurationField(
                        blank=True, db_column="time_to_internal_response", null=True
                    ),
                ),
                (
                    "time_to_resolve",
                    models.DurationField(
                        blank=True, db_column="time_to_resolve", null=True
                    ),
                ),
                (
                    "time_to_repair",
                    models.DurationField(
                        blank=True, db_column="time_to_repair", null=True
                    ),
                ),
                (
                    "ignore",
                    models.BooleanField(
                        default=False,
                        help_text="An ignored incident can be closed right away. In the future, it won't affect statistics or appear in reports. This is useful for duplicates incidents, or false-positives.",
                    ),
                ),
                (
                    "time_to_fix",
                    models.DurationField(
                        blank=True, db_column="time_to_fix", null=True
                    ),
                ),
                (
                    "time_to_recovery",
                    models.DurationField(
                        blank=True, db_column="time_to_recovery", null=True
                    ),
                ),
                (
                    "time_to_respond",
                    models.DurationField(
                        blank=True, db_column="time_to_respond", null=True
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        blank=True,
                        help_text="A comma-separated list of tags.",
                        through="taggit.TaggedItem",
                        to="taggit.Tag",
                        verbose_name="Tags",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="component",
            name="group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="incidents.group"
            ),
        ),
        migrations.AlterField(
            model_name="component",
            name="description",
            field=models.TextField(blank=True),
        ),
        migrations.RemoveField(
            model_name="component",
            name="cachet_id",
        ),
        migrations.RenameField(
            model_name="component",
            old_name="slack_groups",
            new_name="slack_conversations",
        ),
        migrations.RemoveField(
            model_name="component",
            name="on_call",
        ),
        migrations.RemoveField(
            model_name="component",
            name="slack_conversations",
        ),
        migrations.CreateModel(
            name="IncidentUpdate",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("message", models.TextField(blank=True, null=True)),
                (
                    "_status",
                    models.IntegerField(
                        blank=True,
                        choices=[
                            (10, "Open"),
                            (20, "Investigating"),
                            (30, "Mitigating"),
                            (40, "Mitigated"),
                            (50, "Post-mortem"),
                            (60, "Closed"),
                        ],
                        db_column="status",
                        null=True,
                        verbose_name="Status",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "commander",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="commander_for",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "communication_lead",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="communication_lead_for",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="incidents.incident",
                    ),
                ),
                (
                    "severity",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=models.SET(
                            firefighter.incidents.models.severity.Severity.get_default
                        ),
                        to="incidents.severity",
                    ),
                ),
                (
                    "component",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="incidents.component",
                    ),
                ),
                ("description", models.TextField(blank=True, null=True)),
                ("title", models.CharField(blank=True, max_length=128, null=True)),
                ("event_ts", models.DateTimeField()),
                ("event_type", models.CharField(blank=True, max_length=64, null=True)),
            ],
            options={
                "ordering": ["-event_ts"],
            },
        ),
        migrations.AlterField(
            model_name="incident",
            name="component",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="incidents.component"
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="incidents_created_by",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="environment",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="incidents.environment"
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="severity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="incidents.severity"
            ),
        ),
        migrations.CreateModel(
            name="IncidentMembership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="incidents.incident",
                    ),
                ),
                (
                    "inviter",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="members",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="incident",
            name="members",
            field=models.ManyToManyField(
                through="incidents.IncidentMembership", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="commander",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="incidents_commander",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="communication_lead",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="incidents_communication_lead",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="created_by",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="incidents_created_by",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="incident",
            name="private",
            field=models.BooleanField(
                default=False,
                help_text="A private incident is not communicated in #tech-incidents, and its created conversation is private. In the future, we may restrict the visibility to incident members only.",
            ),
        ),
        migrations.AlterField(
            model_name="component",
            name="order",
            field=models.IntegerField(
                default=0,
                help_text="Order of the component in the list. Should be unique per `Group`.",
            ),
        ),
        migrations.AddField(
            model_name="component",
            name="private",
            field=models.BooleanField(
                default=False,
                help_text="If true, incident created with this component won't be communicated, and conversations will be made private. This is useful for sensitive components. In the future, private incidents may be visible only to its members.",
            ),
        ),
        migrations.AddConstraint(
            model_name="incident",
            constraint=models.CheckConstraint(
                check=models.Q(("_status__in", [10, 20, 30, 40, 50, 60])),
                name="incidents_incident__status_valid",
            ),
        ),
        migrations.AddConstraint(
            model_name="incidentupdate",
            constraint=models.CheckConstraint(
                check=models.Q(("_status__in", [10, 20, 30, 40, 50, 60])),
                name="incidents_incidentupdate__status_valid",
            ),
        ),
        migrations.CreateModel(
            name="IncidentCostType",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("category", models.CharField(blank=True, max_length=255, null=True)),
                ("default", models.BooleanField(default=False)),
                ("editable", models.BooleanField(default=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Incident cost type",
                "verbose_name_plural": "Incident cost types",
            },
        ),
        migrations.CreateModel(
            name="IncidentCost",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "amount",
                    models.DecimalField(
                        blank=True,
                        decimal_places=2,
                        help_text="Amount in EUR",
                        max_digits=10,
                        null=True,
                    ),
                ),
                ("details", models.TextField(blank=True, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "cost_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="incident_cost_set",
                        to="incidents.incidentcosttype",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="incident_cost_set",
                        to="incidents.incident",
                    ),
                ),
            ],
            options={
                "verbose_name": "Incident cost",
                "verbose_name_plural": "Incident costs",
            },
        ),
        migrations.AddField(
            model_name="user",
            name="bot",
            field=models.BooleanField(
                default=False,
                help_text="Used for non-person accounts (e.g. integrations that have bot/shared accounts.",
            ),
        ),
        migrations.AddField(
            model_name="component",
            name="deploy_warning",
            field=models.BooleanField(
                default=True,
                help_text="If true, a warning will be sent when creating an incident of high severity with this component.",
            ),
        ),
        migrations.AddField(
            model_name="user",
            name="avatar",
            field=models.URLField(
                blank=True,
                help_text="URL to user avatar. Can be hosted externally.",
                max_length=1024,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="severity",
            name="enabled_create",
            field=models.BooleanField(
                default=True,
                help_text="Can you create a new incident with this severity?",
            ),
        ),
        migrations.AddField(
            model_name="severity",
            name="enabled_update",
            field=models.BooleanField(
                default=True, help_text="Can you update an incident to this severity?"
            ),
        ),
        migrations.AlterField(
            model_name="severity",
            name="order",
            field=models.IntegerField(default=0, unique=True),
        ),
        migrations.AddConstraint(
            model_name="severity",
            constraint=models.UniqueConstraint(
                condition=models.Q(("default", True)),
                fields=("default",),
                name="unique__default",
            ),
        ),
        migrations.CreateModel(
            name="Priority",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=128, unique=True)),
                ("value", models.IntegerField(unique=True)),
                ("emoji", models.CharField(default="🔥", max_length=5)),
                ("description", models.CharField(max_length=128)),
                ("order", models.IntegerField(default=0, unique=True)),
                ("default", models.BooleanField(default=False)),
                ("needs_postmortem", models.BooleanField(default=False)),
                (
                    "enabled_create",
                    models.BooleanField(
                        default=True,
                        help_text="Can you create a new incident with this priority?",
                    ),
                ),
                (
                    "enabled_update",
                    models.BooleanField(
                        default=True,
                        help_text="Can you update an incident to this priority?",
                    ),
                ),
                (
                    "reminder_time",
                    models.DurationField(
                        default=datetime.timedelta(seconds=3600),
                        help_text="Time before an incident without any incident update will receive a reminder.",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name_plural": "priorities",
                "ordering": ["order"],
            },
        ),
        migrations.AddConstraint(
            model_name="priority",
            constraint=models.UniqueConstraint(
                condition=models.Q(("default", True)),
                fields=("default",),
                name="unique_priority_default",
            ),
        ),
        migrations.AlterModelOptions(
            name="severity",
            options={
                "ordering": ["order"],
                "verbose_name": "severity (legacy)",
                "verbose_name_plural": "severities (legacy)",
            },
        ),
        migrations.AddField(
            model_name="priority",
            name="sla",
            field=models.DurationField(
                default=datetime.timedelta(seconds=3600),
                help_text="Service Level Agreement",
                verbose_name="SLA",
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="severity",
            field=models.ForeignKey(
                help_text="Severity (legacy, superseded by priority)",
                on_delete=django.db.models.deletion.PROTECT,
                to="incidents.severity",
            ),
        ),
        migrations.AddField(
            model_name="incidentupdate",
            name="priority",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.SET(
                    firefighter.incidents.models.priority.Priority.get_default
                ),
                to="incidents.priority",
            ),
        ),
        migrations.AlterField(
            model_name="incidentupdate",
            name="severity",
            field=models.ForeignKey(
                blank=True,
                help_text="Superseded by priority",
                null=True,
                on_delete=models.SET(
                    firefighter.incidents.models.severity.Severity.get_default
                ),
                to="incidents.severity",
            ),
        ),
        migrations.AddField(
            model_name="incident",
            name="priority",
            field=models.ForeignKey(
                help_text="Priority",
                on_delete=django.db.models.deletion.PROTECT,
                to="incidents.priority",
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="severity",
            field=models.ForeignKey(
                blank=True,
                help_text="Severity (legacy, superseded by priority)",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="incidents.severity",
            ),
        ),
        migrations.CreateModel(
            name="Impact",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "value",
                    models.CharField(
                        choices=[
                            ("NA", "❔ N/A"),
                            ("HI", "🔼 High"),
                            ("MD", "➡️ Medium"),
                            ("LO", "🔽 Low"),
                            ("NO", "0️⃣ None"),
                        ],
                        max_length=2,
                    ),
                ),
                ("details", models.TextField(blank=True, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="ImpactType",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=64)),
                ("help_text", models.CharField(max_length=128)),
                ("value", models.SlugField(unique=True)),
            ],
            options={
                "verbose_name_plural": "Impact Types",
            },
        ),
        migrations.CreateModel(
            name="IncidentImpact",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "impact",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="incidents.impact",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="incidents.incident",
                    ),
                ),
            ],
            options={
                "verbose_name": "Incident impact",
                "verbose_name_plural": "Incident impacts",
            },
        ),
        migrations.AddField(
            model_name="impact",
            name="impact_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="incidents.impacttype"
            ),
        ),
        migrations.AddField(
            model_name="incident",
            name="impacts",
            field=models.ManyToManyField(
                through="incidents.IncidentImpact", to="incidents.impact"
            ),
        ),
        migrations.AddConstraint(
            model_name="impact",
            constraint=models.CheckConstraint(
                check=models.Q(("value__in", ["NA", "HI", "MD", "LO", "NO"])),
                name="incidents_impact_value_valid",
            ),
        ),
        migrations.AddField(
            model_name="priority",
            name="recommended_response_type",
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
        migrations.CreateModel(
            name="ImpactLevel",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Description for the impact level for this impact type.",
                        max_length=75,
                        null=True,
                    ),
                ),
                (
                    "value",
                    models.CharField(
                        choices=[
                            ("HI", "🔼 High"),
                            ("MD", "➡️ Medium"),
                            ("LO", "🔽 Low"),
                            ("NO", "0️⃣ N/A"),
                        ],
                        default="NO",
                        max_length=2,
                    ),
                ),
                ("order", models.PositiveSmallIntegerField(default=10)),
            ],
        ),
        migrations.RemoveConstraint(
            model_name="impact",
            name="incidents_impact_value_valid",
        ),
        migrations.RemoveField(
            model_name="impact",
            name="value",
        ),
        migrations.AddField(
            model_name="impacttype",
            name="order",
            field=models.PositiveSmallIntegerField(default=10),
        ),
        migrations.AddField(
            model_name="impactlevel",
            name="impact_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="levels",
                to="incidents.impacttype",
            ),
        ),
        migrations.AddField(
            model_name="impact",
            name="impact_level",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="incidents.impactlevel"
            ),
            preserve_default=False,
        ),
        migrations.AddConstraint(
            model_name="impactlevel",
            constraint=models.UniqueConstraint(
                fields=("impact_type", "value"), name="unique_impact_level"
            ),
        ),
        migrations.AddConstraint(
            model_name="impactlevel",
            constraint=models.CheckConstraint(
                check=models.Q(("value__in", ["HI", "MD", "LO", "NO"])),
                name="incidents_impactlevel_value_valid",
            ),
        ),
        migrations.AlterField(
            model_name="impactlevel",
            name="id",
            field=models.UUIDField(editable=False, primary_key=True, serialize=False),
        ),
        migrations.AddField(
            model_name="impactlevel",
            name="emoji",
            field=models.CharField(default="ℹ️", max_length=5),  # noqa: RUF001
        ),
        migrations.AddField(
            model_name="impacttype",
            name="emoji",
            field=models.CharField(default="ℹ️", max_length=5),  # noqa: RUF001
        ),
        migrations.CreateModel(
            name="MilestoneType",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Title or name (e.g. Recovered)", max_length=50
                    ),
                ),
                (
                    "event_type",
                    models.SlugField(
                        help_text="Technical name, slugified version of the name (e.g. recovered)",
                        unique=True,
                    ),
                ),
                (
                    "summary",
                    models.CharField(
                        help_text="Slack Markdown summary of the milestone (e.g. *Recovered*, when the impact is no longer apparent.)",
                        max_length=250,
                    ),
                ),
                (
                    "user_editable",
                    models.BooleanField(
                        help_text="Can be edited by users (else, system event)."
                    ),
                ),
                (
                    "required",
                    models.BooleanField(
                        help_text="Is required for the incident to be considered as complete."
                    ),
                ),
                (
                    "asked_for",
                    models.BooleanField(
                        default=True, help_text="Asked for in the form."
                    ),
                ),
                ("description", models.TextField(blank=True, null=True)),
            ],
            options={
                "verbose_name": "Milestone Type",
                "verbose_name_plural": "Milestone Types",
            },
        ),
        migrations.CreateModel(
            name="MetricType",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Title or name (e.g. Time to Detect)", max_length=50
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        help_text="Short code (e.g. TTD)", max_length=10, unique=True
                    ),
                ),
                ("description", models.TextField()),
                (
                    "type",
                    models.SlugField(
                        blank=True,
                        help_text="Technical name, slugified version of the title (e.g. time_to_detect)",
                        unique=True,
                    ),
                ),
                (
                    "milestone_lhs",
                    models.ForeignKey(
                        help_text="The milestone that ends the metric (duration = milestone_lhs - milestone_rhs)",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metric_type_lhs",
                        to="incidents.milestonetype",
                    ),
                ),
                (
                    "milestone_rhs",
                    models.ForeignKey(
                        help_text="The milestone that starts the metric (duration = milestone_lhs - milestone_rhs)",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metric_type_rhs",
                        to="incidents.milestonetype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Metric Type",
                "verbose_name_plural": "Metric Types",
            },
        ),
        migrations.CreateModel(
            name="IncidentMetric",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "duration",
                    models.DurationField(
                        validators=[
                            django.core.validators.MinValueValidator(
                                datetime.timedelta(0)
                            )
                        ]
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="metric_set",
                        to="incidents.incident",
                    ),
                ),
                (
                    "metric_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="incidents.metrictype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Incident Metric",
                "verbose_name_plural": "Incident Metrics",
            },
        ),
        migrations.AddConstraint(
            model_name="incidentmetric",
            constraint=models.UniqueConstraint(
                fields=("incident", "metric_type"),
                name="incidents_incidentmetric__incident_metric_type_unique",
            ),
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_detect",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_fix",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_internal_response",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_recovery",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_repair",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_resolve",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="time_to_respond",
        ),
        migrations.AddField(
            model_name="incident",
            name="metrics",
            field=models.ManyToManyField(
                through="incidents.IncidentMetric", to="incidents.metrictype"
            ),
        ),
        migrations.AddConstraint(
            model_name="incidentmetric",
            constraint=models.CheckConstraint(
                check=models.Q(("duration__gte", datetime.timedelta(0))),
                name="incidents_incidentmetric__duration_positive",
            ),
        ),
        migrations.AddConstraint(
            model_name="metrictype",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("milestone_lhs", models.F("milestone_rhs")), _negated=True
                ),
                name="incidents_metrictype__milestone_lhs_not_equal_milestone_rhs",
                violation_error_message="The milestone that ends the metric cannot be the same as the milestone that starts the metric.",
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="commander",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="incidents_commander",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="incident",
            name="communication_lead",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="incidents_communication_lead",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="IncidentRoleType",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("slug", models.SlugField(max_length=255, unique=True)),
                (
                    "emoji",
                    models.CharField(
                        blank=True,
                        help_text="The emoji to represent this role.",
                        max_length=5,
                    ),
                ),
                ("summary", models.TextField()),
                ("description", models.TextField()),
                ("aka", models.TextField(blank=True)),
                ("responsibilities", models.TextField(blank=True)),
                ("requirements", models.TextField(blank=True)),
                ("best_practices", models.TextField(blank=True)),
                (
                    "required",
                    models.BooleanField(
                        default=False,
                        help_text="If the role is required, it will be automatically attributed to the incident creator.",
                    ),
                ),
                (
                    "order",
                    models.PositiveSmallIntegerField(
                        default=0, help_text="The display order of the role."
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="IncidentRole",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
        ),
        migrations.RemoveField(
            model_name="incidentmembership",
            name="inviter",
        ),
        migrations.AddConstraint(
            model_name="incidentmembership",
            constraint=models.UniqueConstraint(
                fields=("incident", "user"), name="unique_incident_membership"
            ),
        ),
        migrations.AddField(
            model_name="incidentrole",
            name="incident",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles",
                to="incidents.incident",
            ),
        ),
        migrations.AddField(
            model_name="incidentrole",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="incidentrole",
            name="role",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles",
                to="incidents.incidentroletype",
            ),
        ),
        migrations.AddConstraint(
            model_name="incidentrole",
            constraint=models.UniqueConstraint(
                fields=("incident", "user", "role"), name="unique_incident_role"
            ),
        ),
        migrations.AddField(
            model_name="incident",
            name="roles",
            field=models.ManyToManyField(
                related_name="incident_roles_set",
                through="incidents.IncidentRole",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="incidentrole",
            name="incident",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles_set",
                to="incidents.incident",
            ),
        ),
        migrations.AlterField(
            model_name="incidentrole",
            name="role",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles_set",
                to="incidents.incidentroletype",
            ),
        ),
        migrations.AlterField(
            model_name="incidentrole",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="roles_set",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="impactlevel",
            name="value",
            field=models.CharField(
                choices=[
                    ("HI", "High"),
                    ("MD", "Medium"),
                    ("LO", "Low"),
                    ("NO", "N/A"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
        migrations.RemoveField(
            model_name="incident",
            name="commander",
        ),
        migrations.RemoveField(
            model_name="incident",
            name="communication_lead",
        ),
        migrations.AddConstraint(
            model_name="incidentroletype",
            constraint=models.CheckConstraint(
                check=models.Q(("slug__regex", "^[a-z0-9\\_]+$")),
                name="incidents_incidentroletype_slug_check",
            ),
        ),
        migrations.RemoveField(
            model_name="incidentroletype",
            name="best_practices",
        ),
        migrations.RemoveField(
            model_name="incidentroletype",
            name="requirements",
        ),
        migrations.RemoveField(
            model_name="incidentroletype",
            name="responsibilities",
        ),
        migrations.AddField(
            model_name="incidentroletype",
            name="summary_first_person",
            field=models.TextField(
                blank=True,
                help_text="First person version of the summary, used for instance in Slack messages to assignees.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="incidentrole",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="incidentrole",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.RemoveConstraint(
            model_name="incidentrole",
            name="unique_incident_role",
        ),
        migrations.RenameField(
            model_name="incidentrole",
            old_name="role",
            new_name="role_type",
        ),
        migrations.AddConstraint(
            model_name="incidentrole",
            constraint=models.UniqueConstraint(
                fields=("incident", "user", "role_type"), name="unique_incident_role"
            ),
        ),
    ]



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.utils import timezone
from slack_sdk.errors import SlackApiError, SlackRequestError
from slack_sdk.models.blocks.basic_components import MarkdownTextObject, Option
from slack_sdk.models.blocks.block_elements import ButtonElement, OverflowMenuElement
from slack_sdk.models.blocks.blocks import (
    ActionsBlock,
    Block,
    DividerBlock,
    HeaderBlock,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.firefighter.utils import get_first_in, get_in
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models.incident import Incident
from firefighter.slack.slack_app import DefaultWebClient, SlackApp
from firefighter.slack.slack_templating import (
    date_time,
    slack_block_footer,
    slack_block_help_commands,
    slack_block_help_description,
    slack_block_help_tip,
    slack_block_quote,
)
from firefighter.slack.views.modals.close import CloseModal
from firefighter.slack.views.modals.open import OpenModal
from firefighter.slack.views.modals.update import UpdateModal
from firefighter.slack.views.modals.update_status import modal_update_status

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack
    from slack_sdk.web.client import WebClient

APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME
SLACK_APP_EMOJI: str = settings.SLACK_APP_EMOJI

app = SlackApp()

logger = logging.getLogger(__name__)


@app.event("app_home_opened")
def update_home_tab(
    event: dict[str, Any], client: WebClient = DefaultWebClient
) -> None:
    logger.debug(event)
    # Show only the latest 30 incidents, as Slack does not allow more than 100 elements
    shown_incidents = list(
        Incident.objects.filter(_status__lt=IncidentStatus.CLOSED.value)
        .order_by("-id")
        .select_related(
            "priority", "component", "environment", "component__group", "conversation"
        )[:30]
    )
    blocks: list[Block] = [
        HeaderBlock(text=f"{APP_DISPLAY_NAME} - Incident Management"),
        slack_block_help_description(),
        slack_block_help_commands(),
        slack_block_help_tip(),
        DividerBlock(),
        ActionsBlock(
            block_id="home_actions",
            elements=[
                ButtonElement(
                    text="Open incident",
                    value="open_incident_home_button",
                    style="primary",
                    action_id=OpenModal.open_action,
                ),
                ButtonElement(
                    text="Update critical incident",
                    value="update_incident_home_button",
                    style="primary",
                    action_id=UpdateModal.open_action,
                ),
                ButtonElement(
                    text="Close critical incident",
                    value="close_incident_home_button",
                    style="primary",
                    action_id=CloseModal.open_action,
                ),
            ],
        ),
        HeaderBlock(
            text=f"{len(shown_incidents)} critical incidents active at {datetime.now(tz=timezone.get_current_timezone()).strftime('%Y-%m-%d %H:%M:%S')}"
        ),
    ]

    if len(shown_incidents) == 0:
        blocks.extend((
            DividerBlock(),
            SectionBlock(text="No active incidents! Enjoy :tada:"),
        ))
    else:
        for incident in shown_incidents:
            blocks.extend(_home_incident_element(incident))
    blocks.extend((DividerBlock(), slack_block_footer()))
    view = View(type="home", blocks=blocks)

    try:
        client.views_publish(user_id=event["user"], view=view)
    except (SlackApiError, SlackRequestError):
        logger.exception("Error publishing home tab!")


@app.action("app_home_incident_action")
def handle_some_action(ack: Ack, body: dict[str, Any]) -> None:
    logger.debug(body)

    action = get_first_in(
        body.get("actions", ""),
        key=("selected_option", "value"),
        matches=("update_status", "open_link"),
    )

    action_selected = get_in(action, "selected_option.value")
    if action_selected == "update_status":
        modal_update_status.open_modal_aio(ack, body)

    elif action_selected == "open_link":
        ack()

    else:
        logger.warning(f"Unknown action ! {action_selected}")


def _home_incident_element(
    incident: Incident, *, show_actions: bool = True
) -> list[Block]:
    blocks: list[Block] = [
        DividerBlock(),
        SectionBlock(
            block_id=f"app_home_incident_element_{incident.id}",
            text=f"*#{incident.slack_channel_name if incident.slack_channel_name is not None else incident.id}* - *{incident.title}* - {SLACK_APP_EMOJI} <{incident.status_page_url + '?utm_medium=FireFighter+Slack&utm_source=Slack+Home&utm_campaign=Slack+Home+Link'}| Status Page>",
            fields=[
                MarkdownTextObject(
                    text=f":information_source: *Status:* {incident.status.label}"
                ),
                MarkdownTextObject(
                    text=f":rotating_light: *Priority:* {incident.priority.emoji} {incident.priority.name}"
                ),
                MarkdownTextObject(
                    text=f":package: *Component:* {incident.component.group.name} - {incident.component.name}"
                ),
                MarkdownTextObject(
                    text=f":speaking_head_in_silhouette: *Last update:* {date_time(incident.updated_at)}"
                ),
            ],
            accessory=(
                OverflowMenuElement(
                    action_id="app_home_incident_action",
                    options=[
                        Option(text="Update status", value="update_status"),
                        Option(
                            text="See status page",
                            url=incident.status_page_url
                            + "?utm_medium=FireFighter+Slack&utm_source=Slack+Home&utm_campaign=Slack+Home+Button",
                            value="open_link",
                        ),
                    ],
                )
                if show_actions
                else None
            ),
        ),
        slack_block_quote(incident.description, length=1500),
    ]

    return blocks



from __future__ import annotations

import operator
from datetime import datetime

from django.utils import timezone
from factory import Iterator, LazyAttribute, Sequence, SubFactory
from factory.django import DjangoModelFactory
from factory.faker import Faker
from factory.fuzzy import FuzzyChoice, FuzzyDateTime, FuzzyInteger

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models import (
    Component,
    Environment,
    Group,
    Incident,
    Priority,
    User,
)


class UserFactory(DjangoModelFactory[User]):
    class Meta:
        model = User

    name = Faker("name")  # type: ignore[no-untyped-call]
    email = Faker("email")  # type: ignore[no-untyped-call]
    username = Faker("user_name")  # type: ignore[no-untyped-call]


class GroupFactory(DjangoModelFactory[Group]):
    class Meta:
        model = Group

    name = Faker("text", max_nb_chars=20)  # type: ignore[no-untyped-call]
    description = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]


class SeverityFactory(DjangoModelFactory[Priority]):
    class Meta:
        model = Priority

    name = Faker("text", max_nb_chars=20)  # type: ignore[no-untyped-call]
    value = FuzzyInteger(0, 100)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]


class EnvironmentFactory(DjangoModelFactory[Environment]):
    class Meta:
        model = Environment

    name = Faker("text", max_nb_chars=20)  # type: ignore[no-untyped-call]
    description = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]


class ComponentFactory(DjangoModelFactory[Component]):
    class Meta:
        model = Component

    name = Faker("text", max_nb_chars=30)  # type: ignore[no-untyped-call]
    description = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    order = FuzzyInteger(100, 1000)  # type: ignore[no-untyped-call]
    group = SubFactory(GroupFactory)  # type: ignore[no-untyped-call]


class IncidentFactory(DjangoModelFactory[Incident]):
    class Meta:
        model = Incident

    id = Sequence(lambda n: n + 1)  # type: ignore[no-untyped-call]
    title = Faker("text", max_nb_chars=50)  # type: ignore[no-untyped-call]
    description = Faker("text")  # type: ignore[no-untyped-call]
    _status = FuzzyChoice(IncidentStatus.choices, getter=operator.itemgetter(0))  # type: ignore[no-untyped-call]
    created_at = FuzzyDateTime(
        start_dt=datetime(2018, 1, 1, tzinfo=timezone.get_current_timezone()),
        end_dt=datetime(2022, 1, 1, tzinfo=timezone.get_current_timezone()),
    )  # type: ignore[no-untyped-call]
    updated_at = LazyAttribute(
        lambda inc: datetime(
            inc.created_at.year,
            inc.created_at.month,
            inc.created_at.day,
            tzinfo=timezone.get_current_timezone(),
        )
    )  # type: ignore[no-untyped-call]
    component = Iterator(Component.objects.all())  # type: ignore[no-untyped-call]
    priority = Iterator(Priority.objects.all())  # type: ignore[no-untyped-call]
    environment = Iterator(Environment.objects.all())  # type: ignore[no-untyped-call]

    created_by = SubFactory(UserFactory)  # type: ignore[no-untyped-call]



from __future__ import annotations

import io
import logging
from typing import TYPE_CHECKING, Any, Final, cast

from django.conf import settings
from httpx import HTTPError
from jira.exceptions import JIRAError

from firefighter.firefighter.http_client import HttpClient
from firefighter.firefighter.utils import get_in
from firefighter.jira_app.client import JiraClient
from firefighter.raid.models import FeatureTeam

if TYPE_CHECKING:
    from jira import Project

    from firefighter.jira_app.types import WorkflowBuilderResponse
    from firefighter.raid.types import (
        JiraObject,
    )

logger = logging.getLogger(__name__)

RAID_JIRA_PROJECT_KEY: Final[str] = settings.RAID_JIRA_PROJECT_KEY
TOOLBOX_URL: Final[str] = settings.RAID_TOOLBOX_URL
# XXX Do not hardcode this, it should be a setting or fetched from Jira
RAID_JIRA_WORKFLOW_NAME: Final[str] = "Incident workflow - v2023.03.13"
TARGET_STATUS_NAME: Final[str] = "Closed"


class JiraAttachmentError(Exception):
    pass


class RaidJiraClient(JiraClient):
    def create_issue(  # noqa: PLR0912, PLR0913, C901, PLR0917
        self,
        issuetype: str | None,
        summary: str,
        description: str,
        assignee: str | None,
        reporter: str,
        priority: int | None,
        labels: list[str] | None = None,
        zoho_desk_ticket_id: str | int | None = None,
        zendesk_ticket_id: str | int | None = None,
        is_seller_in_golden_list: bool | None = None,  # noqa: FBT001
        is_key_account: bool | None = None,  # noqa: FBT001
        seller_contract_id: int | str | None = None,
        suggested_team_routing: str | None = None,
        business_impact: str | None = None,
        platform: str | None = None,
        area: str | None = None,
        environments: list[str] | None = None,
        project: str | None = None,
    ) -> JiraObject:
        description_addendum: list[str] = []
        extra_args: dict[str, Any] = {}
        if assignee:
            extra_args["assignee"] = {"id": assignee}
        if labels is None:
            labels = [""]
        if priority is None:
            priority_value = "-"
        else:
            if not 1 <= priority <= 4:
                raise ValueError("Priority must be between 1 and 4")
            priority_value = str(priority)
        if area:
            extra_args["customfield_10920"] = str(area)
        if zoho_desk_ticket_id:
            extra_args["customfield_10896"] = str(zoho_desk_ticket_id)
        if zendesk_ticket_id:
            extra_args["customfield_10895"] = str(zendesk_ticket_id)
        if seller_contract_id:
            description_addendum.append(
                f"Seller link to TOOLBOX: {TOOLBOX_URL}?seller_id={seller_contract_id}"
            )
            extra_args["customfield_10908"] = str(seller_contract_id)
        if is_seller_in_golden_list:
            labels = [*labels, "goldenList"]
        if is_key_account:
            labels = [*labels, "keyAccount"]
        if suggested_team_routing and suggested_team_routing != "SBI":
            description_addendum.append(
                f"Suggested team to be routed: *{suggested_team_routing}*"
            )
        if business_impact and business_impact != "N/A":
            if business_impact not in {"Lowest", "Low", "Medium", "High", "Highest"}:
                raise ValueError("Business impact must be Lowest, Low, Medium, High or Highest")
            extra_args["customfield_10936"] = {"value": business_impact}
        if platform:
            platform = platform.replace("platform-", "")
            extra_args["customfield_10201"] = {"value": platform}
        if environments:
            extra_args["customfield_11049"] = [{"value": e} for e in environments]
        if len(description_addendum) > 0:
            description_addendum_str = "\n *Additional Information* \n" + "\n".join(
                description_addendum
            )
            description += description_addendum_str
        if project is None:
            try:
                feature_team = FeatureTeam.objects.get(name=suggested_team_routing)
            except FeatureTeam.DoesNotExist:
                feature_team = None
            project = (
                feature_team.jira_project_key if feature_team else RAID_JIRA_PROJECT_KEY
            )
        issue = self.jira.create_issue(
            project=project,
            summary=summary,
            description=description,
            issuetype={"name": issuetype},
            reporter={"id": reporter},
            customfield_11064={"value": priority_value},
            labels=labels,
            **extra_args,
        )
        return self._jira_object(issue.raw)

    def get_projects(self) -> list[Project]:
        return self.jira.projects()

    @staticmethod
    def add_attachments_to_issue(issue_id: str | int, urls: list[str]) -> None:
        """Add attachments to a Jira issue.

        Args:
            issue_id (str | int): the Jira issue id
            urls (list[str]): list of urls to the attachments

        Raises:
            JiraAttachmentError: if there is an error while adding any attachment
        """
        http_client = HttpClient()
        for i, url in enumerate(urls):
            index = url.rfind(".")
            extension = url[index:]
            try:
                response = http_client.get(url)
                response.raise_for_status()  # Raises an exception if status code is not 2XX

                in_memory_file = io.BytesIO(response.content)

                client.jira.add_attachment(
                    issue=issue_id,
                    attachment=in_memory_file,
                    filename=f"image{i}{extension}",
                )

            except (HTTPError, JIRAError) as err:
                msg = f"Error while adding attachment to issue: {err}"
                raise JiraAttachmentError(msg) from err

    def close_issue(
        self,
        issue_id: str | int,
    ) -> None:
        return self.transition_issue_auto(
            issue_id, TARGET_STATUS_NAME, RAID_JIRA_WORKFLOW_NAME
        )

    def _get_project_config_workflow(
        self, project_key: str = RAID_JIRA_PROJECT_KEY
    ) -> dict[str, Any]:
        return self._get_project_config_workflow_base(
            project_key, RAID_JIRA_WORKFLOW_NAME
        )

    def _get_project_config_workflow_from_builder(self) -> WorkflowBuilderResponse:
        return self._get_project_config_workflow_from_builder_base(
            RAID_JIRA_WORKFLOW_NAME
        )

    @staticmethod
    def _jira_object(issue: dict[str, Any]) -> JiraObject:
        if issue_id := issue.get("id"):
            jira_id = int(cast("str", issue_id))
        else:
            raise TypeError("Jira ID not found")

        jira_key = issue.get("key")
        jira_assignee_id = get_in(issue, ["fields", "assignee", "accountId"])
        jira_reporter_id = get_in(issue, ["fields", "reporter", "accountId"])
        jira_description = get_in(issue, ["fields", "description"])
        jira_summary = get_in(issue, ["fields", "summary"])
        jira_issue_type = get_in(issue, ["fields", "issuetype", "name"])
        # If any field is not string
        if None in {
            jira_reporter_id,
            jira_description,
            jira_summary,
            jira_issue_type,
        }:
            raise TypeError("Jira object has wrong type")

        if jira_key is None:
            raise TypeError("Jira key is None")

        return {
            "id": jira_id,
            "key": jira_key,
            "project_key": str(jira_key.split("-")[0]),
            "assignee_id": jira_assignee_id,
            "reporter_id": jira_reporter_id,
            "description": jira_description,
            "summary": jira_summary,
            "issue_type": jira_issue_type,
            "business_impact": get_in(
                issue, ["fields", "customfield_10936", "value"], default=""
            ),
        }


client = RaidJiraClient()



from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    TypeAlias,
    TypedDict,
    TypeVar,
    cast,
)
from uuid import UUID

from django import forms
from django.db import models
from django.db.models import Model
from django.utils import timezone
from slack_sdk.models.blocks.basic_components import Option, OptionGroup, TextObject
from slack_sdk.models.blocks.block_elements import (
    CheckboxesElement,
    DateTimePickerElement,
    InputInteractiveElement,
    PlainTextInputElement,
    SelectElement,
    UserSelectElement,
)
from slack_sdk.models.blocks.blocks import ActionsBlock, Block, InputBlock, SectionBlock

from firefighter.firefighter.utils import get_in
from firefighter.incidents.forms.utils import EnumChoiceField, GroupedModelChoiceField
from firefighter.incidents.models.user import User
from firefighter.slack.models.user import SlackUser

if TYPE_CHECKING:
    from collections.abc import Callable

logger = logging.getLogger(__name__)
TZ = timezone.get_current_timezone()


class SlackFormAttributesInput(TypedDict, total=False):
    multiline: bool
    placeholder: str


class SlackFormAttributesBlock(TypedDict, total=False):
    hint: str | None


class SlackFormAttributesWidget(TypedDict, total=False):
    label_from_instance: Callable[[Any], str]
    post_block: Block | None
    pre_block: Block | None


class SlackFormAttributes(TypedDict, total=False):
    input: SlackFormAttributesInput
    block: SlackFormAttributesBlock
    widget: SlackFormAttributesWidget


SlackFormAttributesDict: TypeAlias = dict[str, SlackFormAttributes]


T = TypeVar("T", bound=forms.Form)


class SlackForm(Generic[T]):
    slack_fields: SlackFormAttributesDict
    form: T
    form_class: type[T]

    def __init__(
        self,
        form: type[T],
        slack_fields: SlackFormAttributesDict | None = None,
    ):
        self.form_class = form

        if hasattr(form, "slack_fields") and form.slack_fields:  # type: ignore
            self.slack_fields = form.slack_fields  # type: ignore
        else:
            self.slack_fields = slack_fields or {}

    def __call__(self, *args: Any, **kwargs: Any) -> SlackForm[T]:
        self.form = self.form_class(*args, **kwargs)
        return self

    def slack_blocks(
        self,
        block_wrapper: Literal["section_accessory", "input", "action"] = "input",
    ) -> list[Block]:
        """Return the list of blocks from a SlackForm.

        Args:
            block_wrapper (str, optional): Type of blocks destination. Defaults to "input".

        Raises:
            ValueError: Raised if one of the initial value does not match the field type.

        Returns:
            list[Block]: The List of Slack Blocks.
        """
        blocks = []
        for field_name, f in self.form.fields.items():
            # Get the default data, from the form data, the field.initial or the form.initial
            f.initial = self.get_field_initial(field_name, f)

            slack_input_kwargs: dict[str, Any] = {}
            slack_block_kwargs: dict[str, Any] = {}

            # Set Field common args
            slack_block_kwargs["label"] = (f.label or field_name.title())[:2000]
            slack_block_kwargs["hint"] = f.help_text[:2000]
            slack_block_kwargs["optional"] = not f.required

            # Set custom Slack SDK fields
            post_block, pre_block = self._parse_field_slack_args(
                field_name, f, slack_input_kwargs, slack_block_kwargs
            )

            slack_input_element = self._get_input_element(
                f, field_name, slack_input_kwargs
            )
            if slack_input_element is None:
                continue
            blocks += self._wrap_field_in_block(
                block_wrapper,
                field_name,
                f,
                slack_block_kwargs,
                slack_input_element,
                post_block,
                pre_block,
            )
        return blocks

    def get_field_initial(self, field_name: str, f: forms.Field) -> Any:
        initial = None
        if not self.form.is_bound and self.form.data:
            initial = self.form.cleaned_data.get(field_name)
        else:
            initial = self.form.initial.get(field_name, f.initial)

        if callable(f.initial):
            initial = f.initial()
        return initial

    def _get_input_element(
        self, f: forms.Field, field_name: str, slack_input_kwargs: dict[str, Any]
    ) -> InputInteractiveElement | None:
        if isinstance(f.widget, forms.HiddenInput):
            return None
        match f:
            case forms.CharField():
                if not isinstance(f.initial, str) and f.initial is not None:
                    err_msg = f"Initial value for {field_name} is not a string or None"
                    raise ValueError(err_msg)
                return self._process_slack_char_field(field_name, f, slack_input_kwargs)
            case forms.SplitDateTimeField() | forms.DateTimeField():
                if (
                    not isinstance(f.initial, datetime | str | int)
                    and f.initial is not None
                ):
                    err_msg = f"Initial value for {field_name} is not a datetime, string, int or None"
                    raise ValueError(err_msg)
                return self._process_slack_splitdatetime_field(
                    field_name, f, slack_input_kwargs
                )
            case GroupedModelChoiceField():
                if not isinstance(f.initial, models.Model) and f.initial is not None:
                    err_msg = f"Initial value for {field_name} is not a model instance or None"
                    raise ValueError(err_msg)
                return self._process_slack_model_grouped_choice_field(
                    field_name, f, slack_input_kwargs
                )
            case forms.ModelChoiceField() | forms.ChoiceField() | EnumChoiceField():
                if (
                    isinstance(f, forms.ModelChoiceField)
                    and f.queryset
                    and f.queryset.model == User
                    and (isinstance(f.initial, User) or f.initial is None)
                ):
                    return self._process_model_user_field(
                        field_name, f, slack_input_kwargs
                    )

                return self._process_slack_choice_field(
                    field_name, f, slack_input_kwargs
                )
            case forms.BooleanField():
                return self._process_slack_boolean_field(
                    field_name, f, slack_input_kwargs
                )
            case _:
                logger.warning(f"Field {field_name} of type {type(f)} not supported")
                return None

    @staticmethod
    def _wrap_field_in_block(
        block_wrapper: Literal["section_accessory", "input", "action"],
        field_name: str,
        f: forms.Field,
        slack_block_kwargs: dict[str, Any],
        slack_input_element: InputInteractiveElement,
        post_block: Block | None,
        pre_block: Block | None,
    ) -> list[Block]:
        blocks: list[Block] = []
        if pre_block:
            blocks.append(pre_block)
        if f.disabled:
            if f.label and isinstance(f.label, str):
                blocks.append(SectionBlock(text=f.label))
            if isinstance(f.initial, datetime):
                blocks.append(
                    SectionBlock(
                        text=str(
                            f.initial.astimezone(TZ).strftime("%Y-%m-%d %H:%M %Z")
                            if f.initial
                            else "Not set"
                        )
                    )
                )
            else:
                blocks.append(
                    SectionBlock(text=str(f.initial) if f.initial else "Not set")
                )
        elif block_wrapper == "input":
            blocks.append(
                InputBlock(
                    element=slack_input_element,
                    block_id=field_name,
                    label=slack_block_kwargs.pop("label"),
                    **slack_block_kwargs,
                )
            )
        elif block_wrapper == "section_accessory":
            blocks.append(
                SectionBlock(
                    text=f"*{f.label}*\n{f.help_text}",
                    accessory=slack_input_element,
                )
            )
        elif block_wrapper == "action":
            if f.label and isinstance(f.label, str):
                blocks.append(SectionBlock(text=f.label))
            blocks.append(
                ActionsBlock(
                    elements=[slack_input_element],
                    block_id=field_name,
                    **slack_block_kwargs,
                )
            )
        if post_block:
            blocks.append(post_block)
        return blocks

    def _parse_field_slack_args(
        self,
        field_name: str,
        f: forms.Field,
        slack_input_kwargs: dict[str, Any],
        slack_block_kwargs: dict[str, Any],
    ) -> tuple[Block | None, Block | None]:
        post_block = None
        pre_block = None
        if hasattr(self, "slack_fields") and field_name in self.slack_fields:
            attributes = self.slack_fields[field_name]
            if "input" in attributes:
                input_args = attributes["input"]
                slack_input_kwargs.update(dict(input_args.items()))
            if "block" in attributes:
                block_args = attributes["block"]
                slack_block_kwargs.update(dict(block_args.items()))
            # Set Django field attributes
            if "widget" in attributes:
                widget_args = attributes["widget"]
                if "label_from_instance" in widget_args and isinstance(
                    f, forms.ModelChoiceField
                ):
                    f.label_from_instance = widget_args["label_from_instance"]  # type: ignore[method-assign,assignment]
                if "post_block" in widget_args:
                    post_block = widget_args["post_block"]
                if "pre_block" in widget_args:
                    pre_block = widget_args["pre_block"]
        return post_block, pre_block

    @classmethod
    def _process_slack_choice_field(
        cls,
        field_name: str,
        f: forms.ModelChoiceField | forms.ChoiceField,  # type: ignore[type-arg]
        slack_input_kwargs: dict[str, Any],
    ) -> InputInteractiveElement:
        if not isinstance(f, forms.ModelChoiceField | forms.ChoiceField):
            err_msg = f"Field {field_name} is not a ModelChoiceField or ChoiceField"  # type: ignore[unreachable]
            raise TypeError(err_msg)

        if f.initial:
            initial_choice_label: str | None = None
            if isinstance(f, forms.ModelChoiceField):
                initial_choice_label = f.label_from_instance(f.initial)
                initial_choice_value = str(f.initial.pk)

            elif isinstance(f, EnumChoiceField):
                initial_choice_label = f.initial.label
                initial_choice_value = str(f.initial.value)

            elif isinstance(f, forms.ChoiceField):
                initial_choice_label = str(
                    [c[1] for c in f.choices if c[0] == f.initial][0]  # noqa: RUF015
                )
                initial_choice_value = f.initial
            else:
                err_msg = f"Field {field_name} is not a supported ChoiceField"  # type: ignore[unreachable]
                raise TypeError(err_msg)

            slack_input_kwargs["initial_option"] = SafeOption(
                label=initial_choice_label, value=initial_choice_value
            )

        slack_input_kwargs["options"] = [
            SafeOption(label=str(c[1]), value=str(c[0]))
            for c in filter(lambda co: co[0] != "", f.choices)
        ]
        # Add the initial option to the list of options if it's not there
        if (
            "initial_option" in slack_input_kwargs
            and slack_input_kwargs["initial_option"]
            not in slack_input_kwargs["options"]
        ):
            slack_input_kwargs["options"].append(slack_input_kwargs["initial_option"])

        field_name = f"{field_name}___{f.initial}{datetime.now().timestamp()}"  # noqa: DTZ005
        field_name = field_name[:254]
        return SelectElement(action_id=field_name, **slack_input_kwargs)

    @classmethod
    def _process_model_user_field(
        cls,
        field_name: str,
        f: forms.ModelChoiceField,  # type: ignore[type-arg]
        slack_input_kwargs: dict[str, Any],
    ) -> InputInteractiveElement:
        if not isinstance(f, forms.ModelChoiceField):
            err_msg = f"Field {field_name} is not a ModelChoiceField"  # type: ignore[unreachable]
            raise TypeError(err_msg)

        if f.initial:
            initial_user: User | None = SlackUser.objects.add_slack_id_to_user(
                user=f.initial
            )
            initial_user_slack_id = (
                initial_user.slack_user.slack_id
                if initial_user and initial_user.slack_user
                else None
            )
            slack_input_kwargs["initial_user"] = initial_user_slack_id

        field_name = f"{field_name}___{f.initial}{datetime.now().timestamp()}"  # noqa: DTZ005
        field_name = field_name[:254]
        return UserSelectElement(action_id=field_name, **slack_input_kwargs)

    @classmethod
    def _process_slack_model_grouped_choice_field(
        cls,
        field_name: str,
        f: GroupedModelChoiceField,
        slack_input_kwargs: dict[str, Any],
    ) -> SelectElement:
        if not isinstance(f, GroupedModelChoiceField):
            err_msg = f"Field {field_name} is not a GroupedModelChoiceField"  # type: ignore[unreachable]
            raise TypeError(err_msg)

        if f.initial:
            slack_input_kwargs["initial_option"] = SafeOption(
                label=f.label_from_instance(f.initial),
                value=str(f.initial.pk),
            )

        slack_input_kwargs["option_groups"] = []
        for _, (option_group_value, option_group_label) in enumerate(f.choices):
            if option_group_value is None:
                option_group_value = ""  # noqa: PLW2901
                continue

            if isinstance(option_group_label, list | tuple):
                group_name = str(option_group_value)
                choices = option_group_label
                subgroup = [
                    SafeOption(label=str(option_label), value=str(option_value))
                    for option_value, option_label in choices
                ]

                slack_input_kwargs["option_groups"].append(
                    OptionGroup(label=str(group_name), options=subgroup)
                )

        return SelectElement(action_id=field_name, **slack_input_kwargs)

    @classmethod
    def _process_slack_char_field(
        cls,
        field_name: str,
        f: forms.CharField,
        slack_input_kwargs: dict[str, Any],
    ) -> PlainTextInputElement:
        if not isinstance(f, forms.CharField):
            err_msg = f"Field {field_name} is not a CharField"  # type: ignore[unreachable]
            raise TypeError(err_msg)

        if f.max_length:
            slack_input_kwargs["max_length"] = f.max_length
        if f.min_length:
            slack_input_kwargs["min_length"] = f.min_length

        if f.initial:
            slack_input_kwargs["initial_value"] = f.initial

        return PlainTextInputElement(
            action_id=field_name,
            **slack_input_kwargs,
        )

    @classmethod
    def _process_slack_boolean_field(
        cls,
        field_name: str,
        f: forms.BooleanField,
        slack_input_kwargs: dict[str, Any],
    ) -> CheckboxesElement:
        if not isinstance(f, forms.BooleanField):
            err_msg = f"Field {field_name} is not a BooleanField"  # type: ignore[unreachable]
            raise TypeError(err_msg)
        # Make a list with only one option "Yes" that maps to True
        slack_input_kwargs["options"] = [SafeOption(label="Yes", value="True")]
        if f.initial:
            slack_input_kwargs["initial_option"] = SafeOption(label="Yes", value="True")

        return CheckboxesElement(action_id=field_name, **slack_input_kwargs)

    @classmethod
    def _process_slack_splitdatetime_field(
        cls,
        field_name: str,
        f: forms.SplitDateTimeField | forms.DateTimeField,
        slack_input_kwargs: dict[str, Any],
    ) -> DateTimePickerElement:
        if not isinstance(f, forms.SplitDateTimeField | forms.DateTimeField):
            err_msg = f"Field {field_name} is not a SplitDateTimeField or DateTimeField"  # type: ignore[unreachable]
            raise TypeError(err_msg)

        if f.initial:
            slack_input_kwargs["initial_date_time"] = (
                int(f.initial.timestamp()) if isinstance(f.initial, datetime) else None
            )
            field_name = (
                field_name + "___" + str(datetime.now(timezone.utc).timestamp())  # type: ignore[attr-defined]
            )
            field_name = field_name[:254]
        return DateTimePickerElement(
            action_id=field_name,
            **slack_input_kwargs,
        )


def slack_view_submission_to_dict(
    body: dict[str, Any | str],
) -> dict[str, str | Any]:
    """Returns a dict of the form data from a Slack view submission."""
    if body.get("view"):
        path = ["view", "state", "values"]
    elif body.get("type") == "block_actions":
        path = ["state", "values"]
    else:
        logger.warning("Unknown Slack view submission format: %s", body)
        path = ["view", "state", "values"]

    values: dict[str, dict[str, Any]] = get_in(body, path)
    data: dict[str, str | Any] = {}
    if not isinstance(values, dict):
        raise TypeError("Expected a values dict in the body")

    # We expect only one action per input
    # The action_id must be block_id or block_id___{whatever}
    # We support block_id and block_id___{whatever} because Slack won't update the value in the user's form if the action_id is the same
    # Hence, we need to add a unique identifier to the action_id to force Slack to update the value, replacing the input by another one
    for block in values.values():
        if len(block) == 0:
            continue
        action_id: str = next(iter(block.keys()))
        action_id_stripped = action_id.split("___", 1)[0].strip()
        input_field = block.get(action_id)
        if not isinstance(input_field, dict):
            raise TypeError("Expected input_field to be a dict")

        if input_field.get("type") == "plain_text_input":
            data[action_id_stripped] = input_field.get("value")
        elif input_field.get("type") == "datetimepicker":
            data[action_id_stripped] = (
                datetime.fromtimestamp(
                    cast("float", input_field.get("selected_date_time")),
                    tz=TZ,
                )
                if input_field.get("selected_date_time")
                else None
            )
        elif input_field.get("type") == "static_select":
            data[action_id_stripped] = get_in(input_field, ["selected_option", "value"])
        elif input_field.get("type") == "users_select":
            user_id = get_in(
                input_field,
                [
                    "selected_user",
                ],
            )
            user_obj = (
                SlackUser.objects.get_user_by_slack_id(slack_id=user_id)
                if user_id
                else None
            )
            data[action_id_stripped] = user_obj.id if user_obj else None
    return data


class SlackFormJSONEncoder(json.JSONEncoder):
    """JSON encoder that can handle UUIDs and Django models.
    Used to serialize the form data to JSON for Slack modal private_metadata.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, Model):
            return o.pk
        if isinstance(o, UUID):
            return str(o)
        return super().default(o)


class SafeOption(Option):
    """Make sure we are creating valid Option, warn otherwise."""

    def __init__(
        self,
        *,
        value: str,
        label: str | None = None,
        text: str | dict[str, Any] | TextObject | None = None,  # Block Kit
        description: str | dict[str, Any] | TextObject | None = None,
        url: str | None = None,
        **others: dict[str, Any],
    ) -> None:
        if len(value) > 75:
            logger.warning("Option value is too long: %s", value)
            value = value[:75]
        if label and len(label) > 75:
            logger.warning("Option label is too long: %s", label)
            label = label[:74] + "…"
        super().__init__(
            value=value,
            label=label,
            text=text,
            description=description,
            url=url,
            **others,
        )



"""Collection of custom Django template filters. They are added as built-in filters and thus can be used in templates of any app."""

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast

import markdown as md
import nh3
from django.template.defaulttags import register as register_base

if TYPE_CHECKING:
    from collections.abc import Callable

    from django import template

register_global: template.Library = cast("template.Library", register_base)
V = TypeVar("V")


@register_global.filter
def timedelta_chop_microseconds(delta: timedelta) -> timedelta:
    """Removes microseconds from a timedelta object.

    Args:
        delta (timedelta): The timedelta object to remove microseconds from.

    Returns:
        timedelta: A new timedelta object with microseconds removed.
    """
    return delta - timedelta(microseconds=delta.microseconds)


@register_global.filter
def get_item(dictionary: dict[str, Any], key: Any) -> Any:
    """Get the value of a key from a dictionary or object.

    Args:
        dictionary (dict[str, Any]): The dictionary or object to get the value from.
        key (Any): The key to get the value for.

    Returns:
        Any: The value of the key in the dictionary or object.
    """
    if hasattr(dictionary, key):
        return getattr(dictionary, key)
    if hasattr(dictionary, "__getitem__"):
        try:
            return dictionary[key]
        except KeyError:
            pass
    if hasattr(dictionary, "get"):
        return dictionary.get(key)
    if not isinstance(dictionary, dict):
        dictionary = dictionary.__dict__  # type: ignore

    return dictionary.get(key)


@register_global.filter
def apply_filter(
    value: Any,
    fn_holder: dict[Literal["filter_args", "filter"], Callable[..., V]],
) -> V | Any:
    """Applies a filter function to a given value.

    Args:
        value: The value to be filtered.
        fn_holder: A dictionary containing the filter function and its arguments.

    Returns:
        The filtered value.
    """
    fn = fn_holder.get("filter")

    if not fn:
        return value

    args = fn_holder.get("filter_args")
    if args:
        return fn(value, args)
    return fn(value)


@register_global.filter
def readable_time_delta(delta: timedelta) -> str:
    """Format a time delta as a string. Ignore seconds and microseconds
    From https://github.com/wimglenn/readabledelta/blob/master/readabledelta.py (The Unlicense).
    """
    negative = delta < timedelta(0)
    delta = abs(delta)

    # Other allowed values are  "seconds", "milliseconds", "microseconds".
    keys = [
        "weeks",
        "days",
        "hours",
        "minutes",
    ]

    # datetime.timedelta are normalized internally in Python to the units days, seconds, microseconds allowing a unique
    # representation.  This is not the only possible basis; the calculations below rebase onto more human friendly keys
    # noinspection PyDictCreation
    data = {}
    # rebase days onto weeks, days
    data["weeks"], data["days"] = divmod(delta.days, 7)
    # rebase seconds onto hours, minutes, seconds
    data["hours"], data["seconds"] = divmod(delta.seconds, 60 * 60)
    data["minutes"], data["seconds"] = divmod(data["seconds"], 60)
    # rebase microseconds onto milliseconds, microseconds

    if data["weeks"] != 0 or data["days"] != 0:
        keys.remove("minutes")
        keys.remove("hours")

    output = [
        f"{data[k]} {k[:-1] if data[k] == 1 else k}" for k in keys if data[k] != 0
    ]

    if not output:
        result = "an instant"
    elif len(output) == 1:
        [result] = output
    else:
        left, right = output[:-1], output[-1:]
        result = ", ".join(left) + " and " + right[0]

    if negative:
        return "-" + result
    return result


@register_global.filter
def markdown(text: str) -> str:
    """Converts markdown-formatted text to HTML.

    Sanitize the HTML to only allow a subset of tags.

    Args:
        text (str): The markdown-formatted text to convert.

    Returns:
        str: The HTML-formatted text.
    """
    return nh3.clean(md.markdown(text, output_format="html"), tags=nh3.ALLOWED_TAGS)



from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

import pytest
from slack_sdk import WebClient

from firefighter.incidents.factories import UserFactory
from firefighter.incidents.models import Incident, User
from firefighter.slack.factories import (
    IncidentChannelFactory,
    SlackConversationFactory,
    SlackUserFactory,
)
from firefighter.slack.models import SlackUser
from firefighter.slack.models.conversation import Conversation
from firefighter.slack.models.incident_channel import IncidentChannel

# XXX: Deduplicate from other fixtures


# Fixture to create a SlackUser instance
@pytest.fixture
def slack_user_saved() -> SlackUser:
    user = UserFactory.create()
    return SlackUserFactory.create(user=user)


@pytest.fixture
def conversation() -> Conversation:
    return SlackConversationFactory.create()


@pytest.fixture
def user() -> User:
    return UserFactory()


@pytest.fixture
def slack_user(user: User) -> SlackUser:
    return SlackUserFactory.create(user=user)


class MockWebClient(WebClient):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["token"] = "xoxb-1234-fake-token-test"  # noqa: S105
        kwargs["base_url"] = "https://fake-slack-web-client.localhost"
        super().__init__(*args, **kwargs)
        self.base_url = kwargs["base_url"]
        self.token = kwargs["token"]
        self.conversations_invite = MagicMock()  # type: ignore[method-assign]
        self.conversations_rename = MagicMock()  # type: ignore[method-assign]
        self.users_info = MagicMock()  # type: ignore[method-assign]
        self.conversations_setTopic = MagicMock()  # type: ignore[method-assign]
        self.conversations_join = MagicMock()  # type: ignore[method-assign]


@pytest.fixture
def mock_web_client() -> MockWebClient:
    return MockWebClient()


@pytest.fixture
def incident_channel(incident_saved: Incident) -> IncidentChannel:
    return IncidentChannelFactory.create(incident=incident_saved)



from __future__ import annotations

from firefighter.firefighter.settings.components.common import INSTALLED_APPS
from firefighter.firefighter.settings.settings_utils import config

ENABLE_RAID: bool = config("ENABLE_RAID", cast=bool, default=False)
"Enable the Raid app. Jira app must be enabled and configured as well."

if ENABLE_RAID:
    INSTALLED_APPS += ("firefighter.raid",)

    RAID_DEFAULT_JIRA_QRAFT_USER_ID: str = config("RAID_DEFAULT_JIRA_QRAFT_USER_ID")
    "The default Jira user ID to use for creating issues"

    RAID_JIRA_PROJECT_KEY: str = config("RAID_JIRA_PROJECT_KEY")
    "The Jira project key to use for creating issues, e.g. 'INC'"

    RAID_JIRA_USER_IDS: dict[str, str] = {}
    "Mapping of domain to default Jira user ID"

    RAID_TOOLBOX_URL: str = config("RAID_TOOLBOX_URL")
    "Toolbox URL"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django.conf import settings

from firefighter.jira_app.client import (
    JiraAPIError,
    JiraUserNotFoundError,
)
from firefighter.jira_app.models import JiraUser
from firefighter.raid.client import client as jira_client

if TYPE_CHECKING:
    from firefighter.incidents.models.user import User
    from firefighter.raid.types import JiraObject


logger = logging.getLogger(__name__)
RAID_DEFAULT_JIRA_QRAFT_USER_ID: str = settings.RAID_DEFAULT_JIRA_QRAFT_USER_ID
error_jira_ticket_creation = "Could not create Jira ticket"


def check_issue_id(issue: JiraObject, title: str, reporter: str) -> int | str:
    issue_id = issue.get("id")
    if issue_id is None:
        logger.error(
            f"Could not create Jira ticket for the incident {title} and the reporter {reporter}"
        )
        raise JiraAPIError(error_jira_ticket_creation)
    return issue_id


def get_jira_user_from_user(user: User) -> JiraUser:
    """Returns the JiraUser object for a given user, if it exists in DB or can be fetched on Jira API. Returns the default JiraUser if not found."""
    try:
        jira_user = jira_client.get_jira_user_from_user(user)
    except JiraAPIError:
        logger.exception(f"Could not find Jira user for {user.id}")

        logger.warning(f"User {user.id} has no Jira user")
        try:
            jira_user = jira_client.get_jira_user_from_jira_id(
                RAID_DEFAULT_JIRA_QRAFT_USER_ID
            )
        except JiraUserNotFoundError:
            logger.exception(
                f"Could not find Jira user with account id {RAID_DEFAULT_JIRA_QRAFT_USER_ID}"
            )
            jira_user = JiraUser.objects.get(id=RAID_DEFAULT_JIRA_QRAFT_USER_ID)
    return jira_user


def create_issue_feature_request(
    title: str,
    description: str,
    reporter: str,
    priority: int | None,
    labels: list[str] | None,
    platform: str,
) -> JiraObject:
    """Creates a Jira issue of type Feature Request.

    Args:
        title (str): Summary of the issue
        description (str): Description of the issue
        reporter (str): Jira account id of the reporter
        priority (int): Priority of the issue
        labels (list[str]): Labels to add to the issue
        platform (str): Platform of the issue
    """
    if labels is None:
        labels = [""]
    if "feature-request" not in labels:
        labels.append("feature-request")
    issue = jira_client.create_issue(
        issuetype="Feature Request",
        summary=title,
        description=description,
        assignee=None,
        reporter=reporter,
        priority=priority,
        labels=labels,
        platform=platform,
    )
    check_issue_id(issue, title=title, reporter=reporter)
    return issue


def create_issue_documentation_request(
    title: str,
    description: str,
    reporter: str,
    priority: int | None,
    labels: list[str] | None,
    platform: str,
) -> JiraObject:
    """Creates a Jira issue of type Documentation/Process Request.

    Args:
        title (str): Summary of the issue
        description (str): Description of the issue
        reporter (str): Jira account id of the reporter
        priority (int): Priority of the issue
        labels (list[str]): Labels to add to the issue
        platform (str): Platform of the issue
    """
    if labels is None:
        labels = [""]
    if "documentation-request" not in labels:
        labels.append("documentation-request")
    issue = jira_client.create_issue(
        issuetype="Documentation/Process Request",
        summary=title,
        description=description,
        assignee=None,
        reporter=reporter,
        priority=priority,
        labels=labels,
        platform=platform,
    )
    check_issue_id(issue, title=title, reporter=reporter)
    return issue


def create_issue_internal(
    title: str,
    description: str,
    reporter: str,
    priority: int | None,
    labels: list[str] | None,
    platform: str,
    business_impact: str | None,
    team_to_be_routed: str | None,
    area: str | None,
) -> JiraObject:
    """Creates a Jira Incident Issue of type Internal.

    Args:
        title (str): Summary of the issue
        description (str): Description of the issue
        reporter (str): Jira account id of the reporter
        priority (int): Priority of the issue
        labels (list[str]): Labels to add to the issue
        platform (str): Platform of the issue
        business_impact (str): Business impact of the issue
        team_to_be_routed (str): Team to be routed
        area (str): Area of the issue
    """
    issue = jira_client.create_issue(
        issuetype="Incident",
        summary=title,
        description=description,
        assignee=None,
        reporter=reporter,
        priority=priority,
        labels=labels,
        platform=platform,
        business_impact=business_impact,
        suggested_team_routing=team_to_be_routed,
        area=area,
    )
    check_issue_id(issue, title=title, reporter=reporter)
    return issue


def create_issue_customer(
    title: str,
    description: str,
    reporter: str,
    priority: int | None,
    labels: list[str] | None,
    platform: str,
    business_impact: str | None,
    team_to_be_routed: str | None,
    area: str | None,
    zendesk_ticket_id: str | None,
) -> JiraObject:
    """Creates a Jira Incident issue of type Customer.

    Args:
        title (str): Summary of the issue
        description (str): Description of the issue
        reporter (str): Jira account id of the reporter
        priority (int): Priority of the issue
        labels (list[str]): Labels to add to the issue
        platform (str): Platform of the issue
        business_impact (str): Business impact of the issue
        team_to_be_routed (str): Team to be routed
        area (str): Area of the issue
        zendesk_ticket_id (str): Zendesk ticket id
    """
    issue = jira_client.create_issue(
        issuetype="Incident",
        summary=title,
        description=description,
        assignee=None,
        reporter=reporter,
        priority=priority,
        labels=labels,
        platform=platform,
        business_impact=business_impact,
        suggested_team_routing=team_to_be_routed,
        area=area,
        zendesk_ticket_id=zendesk_ticket_id,
    )
    check_issue_id(issue, title=title, reporter=reporter)
    return issue


def create_issue_seller(  # noqa: PLR0913, PLR0917
    title: str,
    description: str,
    reporter: str,
    priority: int | None,
    labels: list[str] | None,
    platform: str,
    business_impact: str | None,
    team_to_be_routed: str | None,
    area: str | None,
    seller_contract_id: str | None,
    is_key_account: bool | None,  # noqa: FBT001
    is_seller_in_golden_list: bool | None,  # noqa: FBT001
    zoho_desk_ticket_id: str | None,
) -> JiraObject:
    """Creates a Jira Incident issue of type Seller.

    Args:
        title (str): Summary of the issue
        description (str): Description of the issue
        reporter (str): Jira account id of the reporter
        priority (int): Priority of the issue
        labels (list[str]): Labels to add to the issue
        platform (str): Platform of the issue
        business_impact (str): Business impact of the issue
        team_to_be_routed (str): Team to be routed
        area (str): Area of the issue
        seller_contract_id (str): Seller contract id
        is_key_account (bool): Is key account
        is_seller_in_golden_list (bool): Is seller in golden list
        zoho_desk_ticket_id (str): Zoho desk ticket id
    """
    issue = jira_client.create_issue(
        issuetype="Incident",
        summary=title,
        description=description,
        assignee=None,
        reporter=reporter,
        priority=priority,
        labels=labels,
        platform=platform,
        business_impact=business_impact,
        suggested_team_routing=team_to_be_routed,
        area=area,
        seller_contract_id=seller_contract_id,
        is_key_account=is_key_account,
        is_seller_in_golden_list=is_seller_in_golden_list,
        zoho_desk_ticket_id=zoho_desk_ticket_id,
    )
    check_issue_id(issue, title=title, reporter=reporter)
    return issue



import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def get_new_components() -> dict:
    """
    Returns a dictionary of new components to be created.

    Each entry in the dictionary maps a component name to a tuple containing:
    - group_name: The name of the group the component belongs to.
    - slack_channel: The associated Slack channel for the component.

    Returns:
        dict: A mapping of component names to (group name, slack channel) tuples.
    """
    return {
        "Data Tools": ("Data", "impact-data-tools"),
        "Catalog Access": ("Catalog", "impact-catalog-access"),
    }


def add_new_components(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    Group = apps.get_model("incidents", "Group")
    new_components = get_new_components()

    for name, (group_name, _slack) in new_components.items():
        logger.info(f"Creating new component: '{name}' belonging to group '{group_name}'")
        try:
            group_instance = Group.objects.get(name=group_name)
            new_component = Component(name=name, group=group_instance)
            new_component.save()
        except Exception:
            logger.exception(f"Failed to create new group: '{group_name}'.")


def remove_new_components(apps, schema_editor):
    Component = apps.get_model("incidents", "Component")
    new_component_names = get_new_components().keys()

    for name in new_component_names:
        try:
            component = Component.objects.get(name=name)
            logger.info(f"Removing component: '{name}'")
            component.delete()
        except Exception:
            logger.exception(f"Component '{name}' does not exist, skipping removal.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0012_alter_impactlevel"),
    ]

    operations = [
        migrations.RunPython(add_new_components, remove_new_components),
    ]



from __future__ import annotations

from django.conf import settings

from firefighter.api.views import (
    components,
    environments,
    groups,
    incident_cost_types,
    incident_costs,
    incidents,
    severities,
)



from __future__ import annotations

import logging

from celery import shared_task

from firefighter.incidents.models.user import User
from firefighter.jira_app.client import JiraUserNotFoundError, client
from firefighter.jira_app.models import JiraUser

logger = logging.getLogger(__name__)


def add_jira_slack_relationship(user: User) -> JiraUser | None:
    try:
        jira_api_user = client.get_jira_user_from_user(user)
    except JiraUserNotFoundError:
        logger.warning(f"User not found with Jira API '{user.id}'")
        return None
    jira_user_id = jira_api_user.id

    if jira_api_user is None or jira_user_id == "":
        logger.warning(f"Invalid Jira user ID '{jira_user_id}' for User {user.id}")
        return None
    # XXX Check None and valid

    jira_user, _ = JiraUser.objects.update_or_create(
        user=user, defaults={"id": jira_user_id}
    )
    logger.info(f"Added user {jira_user}")
    return jira_user


@shared_task(name="raid.jira_slack_full")
def add_jira_slack_to_all_full() -> None:
    for user in User.objects.exclude(is_active=False).exclude(username=""):
        add_jira_slack_relationship(user)


@shared_task(name="raid.jira_slack_only_missing")
def add_jira_slack_to_all_only_missing() -> None:
    for user in (
        User.objects.exclude(is_active=False)
        .exclude(username="")
        .filter(jira_user__isnull=True)
    ):
        add_jira_slack_relationship(user)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest
from django.urls import reverse

from firefighter.incidents.factories import IncidentFactory, UserFactory

if TYPE_CHECKING:
    from django.test import Client

    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


@pytest.mark.django_db
def test_incidents_dashboard_unauthorized(client: Client) -> None:
    """This test ensures that the incidents dashboard is not accessible, and that we are redirect to auth."""
    response = client.get(reverse("incidents:dashboard"))

    assert response.status_code == 302


@pytest.mark.django_db
def test_incidents_list_unauthorized(client: Client) -> None:
    """This test ensures that the incidents page list is accessible."""
    response = client.get(reverse("incidents:incident-list"))

    assert response.status_code == 302


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incidents_statistics_unauthorized(client: Client) -> None:
    """This test ensures that the incidents statistics page is accessible."""
    response = client.get(reverse("incidents:incident-statistics"))

    assert response.status_code == 302


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incidents_dashboard_authorized(admin_client: Client, admin_user: User) -> None:
    """This test ensures that the incidents dashboard is accessible for an admin."""
    admin_client.force_login(admin_user)
    response = admin_client.get(reverse("incidents:dashboard"))

    assert response.status_code == 200


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incidents_details(client: Client, admin_user: User) -> None:
    """This test ensures that the incidents dashboard is accessible for an admin."""
    user = UserFactory.create()
    incident = IncidentFactory.create(created_by=user)
    client.force_login(admin_user)
    response = client.get(
        reverse("incidents:incident-detail", kwargs={"incident_id": incident.id})
    )

    assert response.status_code == 200


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incidents_details_unauthorized_404_redirect(client: Client) -> None:
    """This test ensures that the incidents dashboard will return a 404."""
    response = client.get(
        reverse("incidents:incident-detail", kwargs={"incident_id": 119999})
    )

    assert response.status_code == 302


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incidents_details_authorized_404(client: Client, admin_user: User) -> None:
    """This test ensures that the incidents dashboard will return a 404."""
    client.force_login(admin_user)
    response = client.get(
        reverse("incidents:incident-detail", kwargs={"incident_id": 119999})
    )

    assert response.status_code == 404


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incident_create_unauthorized(client: Client) -> None:
    """This test ensures that the incident create page is accessible."""
    response = client.get(reverse("incidents:incident-create"))

    # We expect a redirect to the login page
    assert response.status_code == 302


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_component_list(client: Client) -> None:
    """This test ensures that the component list is accessible."""
    response = client.get(reverse("incidents:component-list"))

    assert response.status_code == 302


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_incident_list(client: Client) -> None:
    """This test ensures that the incident list is accessible."""
    response = client.get(reverse("incidents:incident-list"))

    assert response.status_code == 302



from __future__ import annotations

import logging
from functools import cache, cached_property
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import urljoin

from django.conf import settings
from django.urls import reverse

from firefighter.firefighter.utils import get_in

if TYPE_CHECKING:
    from firefighter.confluence.client import ConfluenceClient
    from firefighter.confluence.utils import ConfluencePage, ConfluencePageId, PageInfo
    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


@cache
def get_confluence_client() -> type[ConfluenceClient]:
    # ruff: noqa: PLC0415
    from firefighter.confluence.client import ConfluenceClient

    return ConfluenceClient


class ConfluenceService:
    """Provide methods to interact with Confluence, without regards to the underlying library or HTTP client used.

    TODO: The service should provide Pythonic errors and return types, and should not leak the underlying library or Confluence errors.
    It is not a priority for now, as the Confluence API is not used in critical paths.
    """

    logger = logging.getLogger(__name__)
    _client: ConfluenceClient | None

    POSTMORTEM_FOLDER_ID: int = int(settings.CONFLUENCE_POSTMORTEM_FOLDER_ID)
    POSTMORTEM_TEMPLATE_PAGE_ID: int = int(
        settings.CONFLUENCE_POSTMORTEM_TEMPLATE_PAGE_ID
    )
    RUNBOOKS_FOLDER_ID: int = int(settings.CONFLUENCE_RUNBOOKS_FOLDER_ID)
    DEFAULT_SPACE: str = settings.CONFLUENCE_POSTMORTEM_SPACE

    base_url: str = settings.CONFLUENCE_URL.removesuffix("/rest/api")
    """Confluence base URL. (with `/wiki`, without `/rest/api`)"""

    def __init__(self) -> None:
        self._client = None

    @cached_property
    def client(self) -> ConfluenceClient:
        if self._client is None:
            self._client = get_confluence_client()(
                settings.CONFLUENCE_URL,
                settings.CONFLUENCE_USERNAME,
                settings.CONFLUENCE_API_KEY,
            )
        return self._client

    def _build_page_from_django_template(
        self, template_name: str, context: dict[str, Any]
    ) -> str:
        """Returns a unicode string, from a Django template (or path) and its arguments, if any."""
        from django.template import loader

        template = loader.get_template(template_name)
        rendered_page = template.render(context)
        logger.debug(rendered_page)
        return rendered_page

    def parse_confluence_page(self, pm: dict[str, Any] | ConfluencePage) -> PageInfo:
        """Helper to parse a Confluence page into a PageInfo object.

        Args:
            pm (dict[str, Any] | ConfluencePage): ConfluencePage from the API, or a dict with the same structure.

        Returns:
            PageInfo: The parsed page details that are present on [firefighter.confluence.models.ConfluencePage][].

        Raises:
            TypeError: If one of page_id, page_title, page_url or page_edit_url are not strings.
        """
        page_url = f'{self.base_url}{get_in(pm, "_links.webui")}'
        page_edit_url = f'{self.base_url}{get_in(pm, "_links.editui")}'

        page_id = str(pm.get("id"))
        page_title = pm.get("title")
        if not isinstance(page_id, str):
            raise TypeError("page_id is not a string")
        if not isinstance(page_title, str):
            raise TypeError("page_title is not a string")
        if not isinstance(page_url, str):
            raise TypeError("page_url is not a string")
        if not isinstance(page_edit_url, str):
            raise TypeError("page_edit_url is not a string")
        return {
            "name": page_title,
            "page_id": page_id,
            "page_url": page_url,
            "page_edit_url": page_edit_url,
        }

    def update_oncall_page(self, users: dict[str, User]) -> bool:
        """Update the Confluence list of On-Call users, if the page needs to be updated.
        Users should have a SlackUser AND a PagerDutyUser associated.

        Args:
            users (list[User]): list of Users to update the page with.

        Returns:
            bool: has the page been updated?
        """
        if not settings.CONFLUENCE_ON_CALL_PAGE_ID:
            logger.info("No Confluence OnCall page ID, skipping.")
            return False

        content = (self.client.get_page(settings.CONFLUENCE_ON_CALL_PAGE_ID)).json()

        page_version = get_in(content, "version.number")
        if not page_version:
            logger.error("No page version in Confluence page! %s", content)
            return False
        page_version += 1
        page_body = self._build_page_from_django_template(
            "oncall_team.xml",
            context={
                "users": users.items(),
                "oncall_page_link": urljoin(
                    settings.BASE_URL, reverse("pagerduty:oncall-list")
                ),
            },
        )
        logger.debug("Confluence OnCall page body: %s", page_body)
        if get_in(content, "body.storage.value") != page_body:
            res = self.client.update_page(
                content.get("id"),
                content.get("type"),
                content.get("title"),
                page_body,
                page_version,
            )
            if res.status_code != 200:
                logger.error("Can't update OnCall page: %s", res.json())
                return False
            return True

        logger.info("Confluence OnCall page is up to date, and was not updated.")
        return False

    def create_postmortem(self, title: str) -> ConfluencePage | None:
        """Create a PostMortem page.

        Args:
            title (str): Title of the PostMortem page.

        Returns:
            None | ConfluencePage: The newly created page, or None if it failed.
        """
        # Get the body from the template page
        postmortem_template = (
            self.client.get_page(self.POSTMORTEM_TEMPLATE_PAGE_ID)
        ).json()
        body = get_in(postmortem_template, "body.storage.value")
        logger.debug(body)

        # Create the postmortem with the template body
        page_created: ConfluencePage = self.create_page(
            title, self.POSTMORTEM_FOLDER_ID, body
        )

        logger.debug("Postmortem page id: %s ", page_created.get("id"))
        if "statusCode" in page_created:
            logger.error(
                f"Can't create postmortem page with title {title}: {page_created}"
            )
            return None
        logger.debug(page_created)
        return page_created

    def get_page_children_pages(
        self, page_id: ConfluencePageId, *args: Any, **kwargs: Any
    ) -> list[ConfluencePage]:
        """Get all children pages of a given page.
        TODO: testing.

        Args:
            page_id (ConfluencePageId): ID of the page to get children pages from.
            *args: Arguments to pass to Confluence.
            **kwargs: Keyword arguments to pass to Confluence.

        Returns:
            list[ConfluencePage]: list of children pages.
        """
        return list(self.client.get_page_children_pages(page_id, *args, **kwargs))

    def get_page_descendant_pages(
        self, page_id: ConfluencePageId, *args: Any, **kwargs: Any
    ) -> list[ConfluencePage]:
        """Get all descendant (nested children) pages of a given page.
        TODO: Pagination support and testing.

        Args:
            page_id (ConfluencePageId): ID of the page to get children pages from.
            *args: Arguments to pass to Confluence.
            **kwargs: Keyword arguments to pass to Confluence.

        Returns:
            list[ConfluencePage]: list of children pages.
        """
        return self.client.get_page_descendant_pages(page_id, *args, **kwargs).json()[
            "results"
        ]

    def get_page(self, page_id: ConfluencePageId) -> ConfluencePage:
        """TODO: Errors.

        Args:
            page_id (ConfluencePageId): ID of the page to get.

        Returns:
            ConfluencePage: The page.
        """
        return self.client.get_page(page_id).json()

    def get_page_with_body_and_version(
        self, page_id: ConfluencePageId
    ) -> ConfluencePage:
        return self.client.get_page_body_and_version(page_id).json()

    def create_page(
        self,
        title: str,
        parent_id: ConfluencePageId,
        body: str,
        space: str | None = None,
    ) -> ConfluencePage:
        """TODO: Errors.

        Args:
            title (str): Title of the page to create.
            parent_id (ConfluencePageId): Where to create the page. Will be a child of this page.
            body (str): Body of the page.
            space (str | None, optional): Space to create the page in. If None, defaults to theService's default space. Defaults to None.

        Returns:
            ConfluencePage: The newly created page.
        """
        if not space:
            space = self.DEFAULT_SPACE
        return self.client.create_page(title, parent_id, space, body).json()

    def move_page(
        self,
        page_id: ConfluencePageId,
        target_page_id: ConfluencePageId,
        position: Literal["before", "after", "append"] = "append",
        *,
        dry_run: bool = False,
    ) -> ConfluencePage | None:
        """Args:
            page_id (ConfluencePageId): ID of the page to move.
            target_page_id (ConfluencePageId): ID of the parent page to move the page in relation to.
            position (Literal["before", "after", "append"], optional): Where should the page be moved, in relation to the target page. Defaults to "append".
            dry_run (bool, optional): If True, will log instead of moving. Defaults to False.

        Returns:
            None | ConfluencePage: The moved page, or None if it failed.
        """
        if dry_run:
            logger.info(
                f"Would have moved {page_id} {'into' if position == 'append' else position} {target_page_id} if not dry run"
            )
            return None

        res = self.client.move_page(page_id, target_page_id, position)
        logger.info(
            f"Moved {page_id} {'into' if position == 'append' else position} {target_page_id}"
        )
        if res.status_code == 200:
            return res.json()
        logger.error(f"Error moving page {page_id} {res.status_code} {res.text}")
        return None

    def sort_pages(
        self,
        page_ids: list[tuple[ConfluencePageId, ConfluencePage]],
        *,
        dry_run: bool = False,
    ) -> None:
        """Sort pages according to the list order.

        **Be careful**:
        - this will move pages around, and will not check if the pages are already in the right order.
        - this will not check that the pages are in the same parent.
        - this will perform n moves, where n is the number of pages in the list.

        Args:
            page_ids (list[tuple[ConfluencePageId, ConfluencePage]]): Sorted list of (PageId, Page) tuples.
            dry_run (bool, optional): If True, will log instead of sorting. Defaults to False.
        """
        for index, page in enumerate(page_ids):
            if index == 0:
                continue

            target_id = page_ids[index - 1][0]
            page_id = page[0]
            self.move_page(
                page_id,
                target_id,
                position="after",
                dry_run=dry_run,
            )


confluence_service = ConfluenceService()
if settings.CONFLUENCE_MOCK_CREATE_POSTMORTEM:
    from random import randint

    fake_id = str(randint(10, 99999))  # nosec: B311 # noqa: S311
    confluence_service.create_postmortem = lambda title, _incident: {  # type: ignore
        "id": fake_id,
        "type": "page",
        "status": "current",
        "title": title,
        "_links": {
            "editui": f"/pages/resumedraft.action?draftId={fake_id}",
            "webui": f"/spaces/{settings.CONFLUENCE_POSTMORTEM_SPACE}/pages/{fake_id}",
            "context": "/wiki",
            "self": f"{settings.CONFLUENCE_URL}/content/{fake_id}",
            "tinyui": "/x/ABCDEF",
            "collection": "/rest/api/content",
            "base": confluence_service.base_url,
        },
    }



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.dispatch.dispatcher import receiver

from firefighter.incidents.signals import incident_closed

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
logger = logging.getLogger(__name__)


@receiver(signal=incident_closed)
# pylint: disable=unused-argument
def incident_closed_slack(sender: Any, incident: Incident, **kwargs: Any) -> bool:
    if not hasattr(incident, "conversation"):
        logger.error(f"No Slack conversation to archive for incident {incident}.")
        return False
    return incident.conversation.archive_channel()
    # TODO Retry



from __future__ import annotations

import logging
import os
from typing import Any

from celery import Celery
from django.conf import settings

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "firefighter.firefighter.settings")


if os.environ.get("DD_TRACE_ENABLED", "false").lower() == "true":
    # Enable Datadog tracing except for dev env
    from ddtrace import patch

    patch(celery=True)


app = Celery("firefighter")


# Load task modules from all registered Django app configs.
# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
app.config_from_object(settings.CELERY_SETTINGS)

# Use our own logger (set by importing the settings)
# Log level is set when launching the worker
app.conf.worker_hijack_root_logger = False
app.autodiscover_tasks()


logger = logging.getLogger(__name__)


@app.task(bind=True)
def debug_task(self: Any) -> None:
    logger.info(f"Request: {self.request!r}")



from __future__ import annotations

import logging
from typing import Any

import httpx
from django.conf import settings

FF_HTTP_CLIENT_ADDITIONAL_HEADERS: dict[str, Any] | None = (
    settings.FF_HTTP_CLIENT_ADDITIONAL_HEADERS
)


class HttpClient:
    """Base class for HTTP clients. Uses [httpx](https://www.python-httpx.org/) under the hood.

    Sets some defaults, including:
    - a timeout of 15 seconds for the connection and 20 seconds for the read, to avoid hanging indefinitely
    - access logging

    Used by [firefighter.confluence.client.ConfluenceClient][].
    """

    _logger = logging.getLogger(__name__)
    _client: httpx.Client

    def __init__(self, client_kwargs: dict[str, Any] | None = None) -> None:
        self._client = httpx.Client(**(client_kwargs or {}))
        self._client.timeout = httpx.Timeout(15, read=20)
        if FF_HTTP_CLIENT_ADDITIONAL_HEADERS:
            self._client.headers = httpx.Headers({
                **self._client.headers,
                **FF_HTTP_CLIENT_ADDITIONAL_HEADERS,
            })

    def call(self, method: str, url: str, **kwargs: Any) -> httpx.Response:
        res: httpx.Response = getattr(self._client, method)(url, **kwargs)
        self._logger.info(
            '"%s %s %s" %s',
            res.request.method,
            res.request.url,
            res.http_version,
            res.status_code,
        )
        return res

    def post(self, url: str, **kwargs: Any) -> httpx.Response:
        return self.call("post", url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> httpx.Response:
        return self.call("put", url, **kwargs)

    def get(self, url: str, **kwargs: Any) -> httpx.Response:
        return self.call("get", url, **kwargs)



from __future__ import annotations

from django import forms

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms.utils import EnumChoiceField, GroupedModelChoiceField
from firefighter.incidents.models import Component, Priority


class UpdateStatusForm(forms.Form):
    message = forms.CharField(
        label="Update message",
        widget=forms.Textarea,
        min_length=10,
        max_length=1200,
        required=False,
    )
    status = EnumChoiceField(
        enum_class=IncidentStatus,
        label="Status",
        choices=IncidentStatus.choices_lt(IncidentStatus.CLOSED),
    )
    priority = forms.ModelChoiceField(
        label="Priority",
        queryset=Priority.objects.filter(enabled_update=True),
    )
    component = GroupedModelChoiceField(
        choices_groupby="group",
        label="Issue category",
        queryset=Component.objects.all()
        .select_related("group")
        .order_by(
            "group__order",
            "name",
        ),
    )



from __future__ import annotations

import logging
from typing import Any

from firefighter.slack.models.conversation import Conversation
from firefighter.slack.slack_app import SlackApp

app = SlackApp()

logger = logging.getLogger(__name__)


@app.event("channel_id_changed")
def channel_id_changed(event: dict[str, Any]) -> None:
    logger.debug(event)
    old_channel_id = event.get("old_channel_id")
    new_channel_id = event.get("new_channel_id")
    if not old_channel_id or not new_channel_id:
        logger.warning(f"Invalid event! {event}")
        return

    Conversation.objects.filter(channel_id=old_channel_id).update(
        channel_id=new_channel_id
    )
